"""Print the size of the data returned by a Morphware query (data reported)."""
import json
import sys

import requests

from telliot_core.queries.morphware import Morphware


def main():
    """
    Print the size of the data returned by a Morphware query
    using Morphare-provided data source endpoint:

    curl --request POST http://167.172.239.133:5000/products-2 -H "Content-Type: application/json" \
    -d '{"provider":"amazon","service":"compute","region":"us-east-1"}'

    Example:
    $ python scripts/morphware_data_sizes.py
    """
    # Retrieve data from source provided by Morphware
    headers = {
        # Already added when you pass json=
        # 'Content-Type': 'application/json',
    }
    json_data = {
        "provider": "amazon",
        "service": "compute",
        "region": "us-east-1",
    }
    rsp = requests.post("http://167.172.239.133:5000/products-2", headers=headers, json=json_data)
    unfiltered_data = json.loads(rsp.text)
    print("Num products:", len(unfiltered_data))

    # Remove unneeded product data
    data = []
    for i in range(len(unfiltered_data)):
        data.append(
            json.dumps(
                {
                    "instanceType": unfiltered_data[i]["Instance Type"],
                    "cudaCores": unfiltered_data[i]["CUDA Cores"],
                    "numCPUs": unfiltered_data[i]["Number of CPUs"],
                    "RAM": unfiltered_data[i]["RAM"],
                    "onDemandPricePerHour": unfiltered_data[i]["On-demand Price per Hour"],
                }
            )
        )
    print("Expample filtered data:")
    print(json.dumps(data[:2], indent=4))

    # Get size of encoded string[]
    q = Morphware(version=1)
    submit_value = q.value_type.encode(data)
    print(f"Size of data being reported: {sys.getsizeof(submit_value)} bytes")
    # print(submit_value.hex())


if __name__ == "__main__":
    main()



import os
from pathlib import Path

from telliot_core.apps.telliot_config import TelliotConfig
from telliot_core.model.endpoints import RPCEndpoint


def prep_dir(clean=False):
    """Prepare temporary test directory"""
    tmpdir = Path("./temp").resolve().absolute()
    tmpdir.mkdir(parents=True, exist_ok=True)

    main_file = tmpdir / "main.yaml"
    ep_file = tmpdir / "endpoints.yaml"
    api_keys_file = tmpdir / "api_keys.yaml"
    chain_file = tmpdir / "chains.json"
    directory_file = tmpdir / "directory.json"

    if clean:
        if main_file.exists():
            os.remove(main_file)
        if ep_file.exists():
            os.remove(ep_file)
        if api_keys_file.exists():
            os.remove(api_keys_file)
        if chain_file.exists():
            os.remove(chain_file)
        if directory_file.exists():
            os.remove(directory_file)

    return tmpdir


def test_telliot_config():
    """Test main telliot_core configuration"""
    tmpdir = prep_dir()

    cfg = TelliotConfig(config_dir=tmpdir)

    ep = cfg.get_endpoint()
    assert isinstance(ep, RPCEndpoint)

    api_key = cfg.api_keys.find("anyblock")[0]
    assert api_key.url == "https://api.anyblock.tools/"

    tmpdir = prep_dir(clean=True)



from datetime import datetime
from datetime import timedelta
from datetime import timezone
from typing import Union


def now() -> datetime:
    """Return now as a UTC timestamp"""
    return datetime.now(timezone.utc)


class TimeStamp:
    """Representation of a timestamp with an integer number of seconds"""

    # Access the underlying unix timestamp
    ts = property(lambda self: self._ts)
    _ts: int

    def __init__(self, ts: [int, float]):
        """Create from unix timestamp"""
        self._ts = round(ts)

    @classmethod
    def now(cls) -> "TimeStamp":
        return cls(datetime.now().timestamp())

    def __repr__(self) -> str:
        return f"TimeStamp({self._ts})"

    def __str__(self) -> str:
        return str(self.dt)

    @property
    def dt(self) -> datetime:
        """Timestamp as a datetime object"""
        return datetime.fromtimestamp(self._ts)

    @property
    def age(self) -> Union[timedelta, datetime]:
        age = datetime.now() - self.dt
        rounded_age = age - timedelta(microseconds=age.microseconds)
        return rounded_age



import logging
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from traceback import format_tb
from typing import Optional
from typing import Union

import aiohttp
from chained_accounts import ChainedAccount
from chained_accounts import find_accounts

from telliot_core.apps.session_manager import ClientSessionManager
from telliot_core.apps.telliot_config import TelliotConfig
from telliot_core.contract.contract import Contract
from telliot_core.contract.listener import Listener
from telliot_core.directory import contract_directory
from telliot_core.logs import init_logging
from telliot_core.model.endpoints import RPCEndpoint
from telliot_core.tellor.tellor360.autopay import Tellor360AutopayContract
from telliot_core.tellor.tellor360.oracle import Tellor360OracleContract
from telliot_core.tellor.tellorflex.autopay import TellorFlexAutopayContract
from telliot_core.tellor.tellorflex.oracle import TellorFlexOracleContract
from telliot_core.tellor.tellorflex.token import TokenContract
from telliot_core.tellor.tellorx.master import TellorxMasterContract
from telliot_core.tellor.tellorx.oracle import TellorxOracleContract
from telliot_core.utils.home import telliot_homedir
from telliot_core.utils.versions import show_telliot_versions

NETWORKS = {
    1: "eth-mainnet",
    3: "eth-ropsten",
    4: "eth-rinkeby",
    5: "eth-goerli",
    137: "polygon-mainnet",
    80001: "polygon-mumbai",
    122: "fuse-mainnet",
    69: "optimism-kovan",
    1666600000: "harmony-mainnet",
    1666700000: "harmony-testnet",
    421611: "arbitrum-rinkeby",
    941: "pulsechain-testnet",
    42161: "arbitrum",
    1337: "brownie-local-network",
}

LOGLEVEL_MAP = {
    "CRITICAL": logging.CRITICAL,
    "ERROR": logging.ERROR,
    "WARNING": logging.WARNING,
    "INFO": logging.INFO,
    "DEBUG": logging.DEBUG,
}


@dataclass
class TellorxContractSet:
    master: TellorxMasterContract
    oracle: TellorxOracleContract
    governance: Contract
    treasury: Contract


@dataclass
class TellorFlexContractSet:
    oracle: TellorFlexOracleContract
    autopay: TellorFlexAutopayContract
    token: TokenContract


@dataclass
class Tellor360ContractSet:
    """Tellor360 contract set"""

    oracle: Tellor360OracleContract
    autopay: Tellor360AutopayContract
    token: TokenContract


class TelliotCore:
    """Telliot core application"""

    #: BaseApplication Name
    name: str = "telliot-core"

    #: Home directory
    homedir = property(lambda self: self._homedir)
    _homedir: Path

    #: BaseApplication configuration object
    config = property(lambda self: self._config)
    _config: TelliotConfig

    def get_tellorflex_contracts(self) -> TellorFlexContractSet:
        """Get or create tellorflex contracts."""
        if not self._tellorflex:
            account = self.get_account()

            oracle = TellorFlexOracleContract(node=self.endpoint, account=account)
            oracle.connect()

            token = TokenContract(node=self.endpoint, account=account)
            token.connect()

            autopay = TellorFlexAutopayContract(node=self.endpoint, account=account)
            autopay.connect()

            self._tellorflex = TellorFlexContractSet(oracle=oracle, token=token, autopay=autopay)

        return self._tellorflex

    _tellorflex: Optional[TellorFlexContractSet]

    def get_tellorx_contracts(self) -> TellorxContractSet:
        """Get or create TellorX contracts"""

        if not self._tellorx:
            account = self.get_account()

            master = TellorxMasterContract(node=self.endpoint, account=account)
            master.connect()

            oracle = TellorxOracleContract(node=self.endpoint, account=account)
            oracle.connect()

            self._tellorx = TellorxContractSet(
                master=master,
                oracle=oracle,
                governance=self.get_contract(name="tellorx-governance"),
                treasury=self.get_contract(name="tellorx-treasury"),
            )

        return self._tellorx

    _tellorx: Optional[TellorxContractSet]

    def get_tellor360_contracts(self) -> Tellor360ContractSet:
        """Get or create Tellor360 contracts"""

        if not self._tellor360:
            account = self.get_account()

            oracle = Tellor360OracleContract(node=self.endpoint, account=account)
            oracle.connect()

            autopay = Tellor360AutopayContract(node=self.endpoint, account=account)
            autopay.connect()

            token = TokenContract(node=self.endpoint, account=account)
            token.connect()

            self._tellor360 = Tellor360ContractSet(oracle=oracle, autopay=autopay, token=token)

        return self._tellor360

    _tellor360: Optional[Tellor360ContractSet]

    #: User-specified account name
    account_name = property(lambda self: self._account_name)
    _account_name: Optional[str] = None

    @property
    def shared_session(self) -> aiohttp.ClientSession:
        """Return the shared session"""
        return self._session_manager.session

    _session_manager: ClientSessionManager

    @property
    def listener(self) -> Listener:
        """Get or create listener object"""
        if not self._listener:
            self._listener = Listener(session=self.shared_session, ws_url=self.endpoint.url)

        return self._listener

    _listener: Optional[Listener]

    @property
    def endpoint(self) -> RPCEndpoint:
        """Get or create the endpoint for the current configuration"""
        if not self._endpoint:
            self._endpoint = self.get_endpoint()
            connected = self._endpoint.connect()
            if not connected:
                raise Exception(f"Could not connect to endpoint: {self._endpoint.url}")

        return self._endpoint

    _endpoint: Optional[RPCEndpoint]

    @property
    def log(self) -> logging.Logger:
        """Provide access to the main telliot logger"""
        return self._log

    def __init__(
        self,
        *,
        homedir: Optional[Union[str, Path]] = None,
        config: Optional[TelliotConfig] = None,
        chain_id: Optional[int] = None,
        account_name: Optional[str] = None,
    ):

        self._homedir = telliot_homedir(homedir)
        self._config = config or TelliotConfig(config_dir=self.homedir)
        self._endpoint = None
        self._listener = None
        self._tellorx = None
        self._tellorflex = None
        self._tellor360 = None

        loglevel = LOGLEVEL_MAP[self._config.main.loglevel]
        self._log = init_logging(loglevel)

        if chain_id is not None:
            # Override chain ID
            self._config.main.chain_id = chain_id

        if account_name is not None:
            self.set_account_name(account_name)

        self._session_manager = ClientSessionManager()

        show_telliot_versions(self.log.info)

    def set_account_name(self, account_name: str) -> None:
        _ = self.get_account(name=account_name)  # Make sure it exists
        self._account_name = account_name

    async def startup(self) -> None:
        """Connect to the tellorX network"""
        assert self.config

        chain_id = self.config.main.chain_id

        account = self.get_account()
        if account is None:
            raise RuntimeError("Cannot start tellor-core application.  No account found.")

        await self._session_manager.open()

        msg = f"Connected to {NETWORKS[chain_id]} [default account: {account.name}], time: {datetime.now()}"
        self.log.info(msg)

    def get_endpoint(
        self,
        *,
        chain_id: Optional[int] = None,
        provider: Optional[str] = None,
    ) -> RPCEndpoint:

        if not chain_id:
            chain_id = self.config.main.chain_id

        endpoints = self.config.endpoints.find(chain_id=chain_id, provider=provider)
        if len(endpoints) == 0:
            raise Exception("No endpoints found")

        return endpoints[0]  # type: ignore

    def get_contract(
        self,
        *,
        org: Optional[str] = None,
        name: Optional[str] = None,
        address: Optional[str] = None,
        chain_id: Optional[int] = None,
        account: Optional[ChainedAccount] = None,
    ) -> Contract:

        assert self.config is not None

        if not chain_id:
            chain_id = self.config.main.chain_id
            assert chain_id is not None

        if not account:
            account = self.get_account()

        entries = contract_directory.find(org=org, name=name, address=address, chain_id=chain_id)
        if len(entries) > 1:
            raise Exception("More than one contract found.")
        elif len(entries) == 0:
            raise Exception("No contracts found")

        contract_info = entries[0]

        contract_abi = contract_info.get_abi(chain_id=chain_id)

        if self.endpoint.chain_id is not chain_id:
            raise Exception(f"Endpoint chain {self.endpoint.chain_id} does not match requested chain {chain_id}")

        contract = Contract(
            address=contract_info.address[chain_id],
            abi=contract_abi,
            node=self.endpoint,
            account=account,
        )
        contract.connect()
        return contract

    def get_account(
        self,
        *,
        name: Optional[str] = None,
        address: Optional[str] = None,
        chain_id: Optional[int] = None,
    ) -> ChainedAccount:
        """Retrieve the user specified account

        If None configured, the default first matching account will be used
        """

        if name:
            acc_name = name
        else:
            acc_name = self.account_name

        if not chain_id:
            chain_id = self.config.main.chain_id

        accounts = find_accounts(
            name=acc_name,
            address=address,
            chain_id=chain_id,
        )

        if len(accounts) > 0:
            return accounts[0]
        else:
            raise Exception("No accounts found")

    async def shutdown(self) -> None:
        """Cleanly shutdown core"""

        # Release/close endpoint
        if self._endpoint:
            self._endpoint = None

        # SHut down listeners
        if self._listener:
            await self._listener.shutdown()

        # Close aiohttp session
        await self._session_manager.close()

    async def __aenter__(self) -> "TelliotCore":
        await self.startup()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:  # type: ignore
        if exc_type:
            self.log.error("Exception occurred in telliot-core app")
            self.log.error(exc_type)
            self.log.error(exc_val)
            self.log.error(format_tb(exc_tb))
        await self.shutdown()



import click
from chained_accounts import ChainedAccount
from chained_accounts import find_accounts

from telliot_core.cli.commands.read import get_staker_info
from telliot_core.cli.utils import async_run
from telliot_core.cli.utils import cli_config


@click.group()
@click.pass_context
@click.option("-n", "--name", required=False, help="ChainedAccount name.")
def account(ctx: click.Context, name: str) -> None:
    """Manage Telliot stakers."""
    ctx.obj["ACCOUNT_NAME"] = name


@account.command()
@click.pass_context
@async_run
async def status(ctx: click.Context) -> None:
    """Get on-chain staker status."""
    cfg = cli_config(ctx)
    account_name = ctx.obj["ACCOUNT_NAME"]
    if account_name:
        account = ChainedAccount.get(account_name)
    else:
        accounts = find_accounts(chain_id=cfg.main.chain_id)
        account = accounts[0]

    account_status, date_staked = await get_staker_info(ctx, account.address)

    if not account_status:
        print("Failed to retrieve account status.  Check command arguments")
        return

    if account_name:
        print(f"Account name: {account_name}")

    print(f"Status: {account_status}")
    if account_status != "NotStaked":
        print(f"Staked on {date_staked} ({date_staked.age} ago)")



import pytest
from brownie import accounts
from brownie import TellorFlex

from telliot_core.apps.core import TelliotCore
from telliot_core.tellor.tellorflex.oracle import TellorFlexOracleContract
from telliot_core.utils.response import ResponseStatus
from telliot_core.utils.timestamp import TimeStamp

# from telliot_core.queries.price.spot_price import SpotPrice


@pytest.fixture(scope="module")
def mock_flex_contract():
    """Mock the TellorFlex contract"""
    return accounts[0].deploy(
        TellorFlex,
        "0x0000000000000000000000000000000000000123",
        "0x0000000000000000000000000000000000000456",
        42e18,
        60 * 60,
    )


@pytest.mark.asyncio
async def test_main(mumbai_test_cfg, mock_flex_contract):
    """Test the TellorFlex contract"""
    async with TelliotCore(config=mumbai_test_cfg) as core:
        account = core.get_account()
        # Override contract addresses with locally deployed mock contract addresses
        oracle = TellorFlexOracleContract(core.endpoint, account)
        oracle.address = mock_flex_contract.address
        oracle.connect()

        governance_address = await oracle.get_governance_address()
        assert governance_address == "0x0000000000000000000000000000000000000456"

        stake_amount = await oracle.get_stake_amount()
        assert stake_amount == 42

        tlnv, status = await oracle.get_time_of_last_new_value()
        assert isinstance(status, ResponseStatus)
        if status.ok:
            assert isinstance(tlnv, TimeStamp)
        else:
            assert tlnv is None

        lock = await oracle.get_reporting_lock()
        assert lock == 60 * 60

        token_address = await oracle.get_token_address()
        assert token_address == "0x0000000000000000000000000000000000000123"

        total_stake = await oracle.get_total_stake_amount()
        assert total_stake == 0

        staker_info, status = await oracle.get_staker_info(core.get_account().address)
        assert isinstance(status, ResponseStatus)
        if status.ok:
            for info in staker_info:
                assert isinstance(info, int)
        else:
            assert staker_info is None

        # q = SpotPrice(asset="btc", currency="USD")
        qid = b"0000000000000000000000000000000000000000000000000000000000000064"
        count, status = await oracle.get_new_value_count_by_qeury_id(qid)

        assert isinstance(status, ResponseStatus)
        if status.ok:
            assert isinstance(count, int)
        else:
            assert count is None



import json
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from typing import Any
from typing import Optional

import requests
from clamfig import deserialize
from clamfig import Serializable

from telliot_core.apps.config import ConfigOptions
from telliot_core.utils.home import TELLIOT_CORE_ROOT


# Read contract ABIs from json files
_abi_folder = Path(__file__).resolve().parent / "data" / "abi"


@dataclass
class ContractInfo(Serializable):
    name: str
    org: str
    address: dict[int, str]
    abi_file: Optional[str] = None

    _abi: Optional[list[Any]] = field(default=None, init=False, repr=False)

    def get_abi(self, chain_id: int = 0, api_key: str = "") -> list[Any]:
        """Returns the contract ABI.

        The ABI is lazily loaded from a file the first time it is requested
        and stored for later access.  If an abi file is not defined, an attempt
        is made to retrieve the ABI from the chain explorer.
        """
        if not chain_id:
            chain_id = list(self.address.keys())[0]

        if not self._abi:
            if self.abi_file:
                with open(_abi_folder / self.abi_file, "r") as f:
                    self._abi = json.load(f)
            else:
                # try to get from etherscan or other explorer using example:
                address = self.address[chain_id]
                if chain_id == 1:
                    url = "https://api.etherscan.io"
                elif chain_id == 3:
                    url = "https://api-ropsten.etherscan.io"
                elif chain_id == 4:
                    url = "https://api-rinkeby.etherscan.io"
                elif chain_id == 42:
                    url = "https://api-kovan.etherscan.io"
                elif chain_id == 137:
                    url = "https://api.polygonscan.com"
                elif chain_id == 80001:
                    url = "https://api-testnet.polygonscan.com"
                elif chain_id == 42161:
                    url = "https://api.arbiscan.io/"
                else:
                    raise ValueError(f"Could not retrieve ABI using chain_id {chain_id}")

                url = url + f"/api?module=contract&action=getabi&address={address}&format=raw"

                if api_key:
                    url = url + f"&apikey={api_key}"

                headers = {"User-Agent": "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:77.0) Gecko/20190101 Firefox/77.0"}
                response = requests.get(url, headers=headers)
                self._abi = response.json()

        return self._abi  # type: ignore

    def restore_state(self, state: dict[Any, Any]) -> None:
        """Workaround JSON dict key type issue.  This should be handled by clamfig in future."""
        strkeys = list(state["address"].keys())
        for chain_id in strkeys:
            state["address"][int(chain_id)] = state["address"].pop(chain_id)

        super().restore_state(state)


@dataclass
class ContractDirectory(ConfigOptions):
    """Contract directory object"""

    entries: dict[str, ContractInfo] = field(default_factory=dict)

    def add_entry(self, entry: ContractInfo) -> None:
        """Add ContractInfo object to directory."""

        if entry.name in self.entries:
            raise ValueError(f"Contrct {entry.name} already in directory")

        self.entries[entry.name] = entry

    @classmethod
    def from_file(cls, filepath: Path) -> "ContractDirectory":
        """Create a ContractDirectory from file."""

        with open(filepath) as f:
            state = json.load(f)

        entry_list = deserialize(state)

        obj = cls(entries={})

        for entry in entry_list:
            obj.add_entry(entry)

        return obj

    def find(
        self,
        *,
        org: Optional[str] = None,
        name: Optional[str] = None,
        address: Optional[str] = None,
        chain_id: Optional[int] = None,
    ) -> list[ContractInfo]:
        """Search the Contract Directory."""

        result = []
        for info in self.entries.values():
            if org is not None:
                if org != info.org:
                    continue
            if chain_id is not None:
                if chain_id not in info.address.keys():
                    continue
            if name is not None:
                if name not in info.name:
                    continue
            if address is not None:
                if address not in info.address.values():
                    continue

            result.append(info)

        return result


contract_directory = ContractDirectory.from_file(TELLIOT_CORE_ROOT / "data/contract_directory.json")



import pytest
from brownie import accounts
from brownie import AutopayMock

from telliot_core.apps.core import TelliotCore
from telliot_core.tellor.tellorflex.autopay import TellorFlexAutopayContract


@pytest.fixture(scope="module")
def mock_autopay_contract():
    return accounts[0].deploy(AutopayMock)


@pytest.mark.asyncio
async def test_get_current_tip(mumbai_test_cfg, mock_autopay_contract):
    async with TelliotCore(config=mumbai_test_cfg) as core:
        account = core.get_account()

        autopay = TellorFlexAutopayContract(core.endpoint, account)
        autopay.address = mock_autopay_contract.address
        autopay.connect()

        query_id = bytes.fromhex("0000000000000000000000000000000000000000000000000000000000000002")

        result, status = await autopay.get_current_tip(query_id=query_id)
        assert status.ok
        assert result == len(query_id)



"""Pytest Fixtures used for testing Pytelliot"""
import os

import pytest
from brownie import chain
from chained_accounts import ChainedAccount
from chained_accounts import find_accounts

from telliot_core.apps.telliot_config import TelliotConfig


@pytest.fixture(scope="session", autouse=True)
def rinkeby_cfg():
    """Get rinkeby endpoint from config

    If environment variables are defined, they will override the values in config files
    """
    cfg = TelliotConfig()

    # Override configuration for rinkeby testnet
    cfg.main.chain_id = 4

    rinkeby_endpoint = cfg.get_endpoint()
    # assert rinkeby_endpoint.network == "rinkeby"

    if os.getenv("NODE_URL", None):
        rinkeby_endpoint.url = os.environ["NODE_URL"]

    rinkeby_accounts = find_accounts(chain_id=4)
    if not rinkeby_accounts:
        # Create a test account using PRIVATE_KEY defined on github.
        key = os.getenv("PRIVATE_KEY", None)
        if key:
            ChainedAccount.add("git-rinkeby-key", chains=4, key=os.environ["PRIVATE_KEY"], password="")
        else:
            raise Exception("Need a rinkeby account")

    return cfg


@pytest.fixture(scope="session", autouse=True)
def mumbai_cfg():
    """Return a test telliot configuration for use on polygon-mumbai

    If environment variables are defined, they will override the values in config files
    """
    cfg = TelliotConfig()

    # Override configuration for mumbai testnet
    cfg.main.chain_id = 80001

    endpt = cfg.get_endpoint()
    if "INFURA_API_KEY" in endpt.url:
        endpt.url = f'https://polygon-mumbai.infura.io/v3/{os.environ["INFURA_API_KEY"]}'

    mumbai_accounts = find_accounts(chain_id=80001)
    if not mumbai_accounts:
        # Create a test account using PRIVATE_KEY defined on github.
        key = os.getenv("PRIVATE_KEY", None)
        if key:
            ChainedAccount.add(
                "git-mumbai-key",
                chains=80001,
                key=os.environ["PRIVATE_KEY"],
                password="",
            )
        else:
            raise Exception("Need a mumbai account")

    return cfg


def local_node_cfg(chain_id: int):
    """Return a test telliot configuration for use of tellorFlex contracts. Overrides
    the default Web3 provider with a local Ganache endpoint.
    """

    cfg = TelliotConfig()

    # Use a chain_id with TellorFlex contracts deployed
    cfg.main.chain_id = chain_id

    endpt = cfg.get_endpoint()

    # Configure testing using local Ganache node
    endpt.url = "http://127.0.0.1:8545"

    # Advance block number to avoid assertion error in endpoint.connect():
    # connected = self._web3.eth.get_block_number() > 1
    chain.mine(10)

    accounts = find_accounts(chain_id=chain_id)
    if not accounts:
        # Create a test account using PRIVATE_KEY defined on github.
        key = os.getenv("PRIVATE_KEY", None)
        if key:
            ChainedAccount.add(
                "git-tellorflex-test-key",
                chains=chain_id,
                key=os.environ["PRIVATE_KEY"],
                password="",
            )
        else:
            raise Exception(f"Need an account for {chain_id}")

    return cfg


@pytest.fixture
def ropsten_test_cfg(scope="session", autouse=True):
    return local_node_cfg(chain_id=3)


@pytest.fixture
def mumbai_test_cfg(scope="session", autouse=True):
    return local_node_cfg(chain_id=80001)


@pytest.fixture
def rinkeby_test_cfg(scope="session", autouse=True):
    return local_node_cfg(chain_id=4)


@pytest.fixture(scope="session", autouse=True)
def fuse_cfg():
    """Return a test telliot configuration for use on polygon-mumbai

    If environment variables are defined, they will override the values in config files
    """
    cfg = TelliotConfig()

    # Override configuration for fuse testnet
    cfg.main.chain_id = 122

    accounts = find_accounts(chain_id=122)
    if not accounts:
        # Create a test account using PRIVATE_KEY defined on github.
        key = os.getenv("PRIVATE_KEY", None)
        if key:
            ChainedAccount.add("git-fuse-key", chains=122, key=os.environ["PRIVATE_KEY"], password="")
        else:
            raise Exception("Need a Fuse account")

    return cfg



""" Telliot application helpers

"""
import logging
import threading
from dataclasses import dataclass
from dataclasses import field
from typing import Optional

logger = logging.getLogger(__name__)


@dataclass
class ThreadedApplication:

    #: BaseApplication Name
    name: str = field(default="appname")

    #: Private thread storage
    _thread: Optional[threading.Thread] = None
    _shutdown: threading.Event = field(default_factory=threading.Event)

    def startup(self) -> None:
        """Startup BaseApplication

        Start the main application thread
        """
        logger.info("Starting {} BaseApplication".format(self.name))
        self._thread = threading.Thread(target=self.main, name=self.name)
        self._thread.start()

    def shutdown(self) -> None:
        """Startup BaseApplication

        Send a shutdown event to the main application thread and wait
        for it to terminate.
        """
        if self._thread:
            logger.info("Terminating {} BaseApplication".format(self.name))

            self._shutdown.set()

            #: Join worker thread
            self._thread.join()

            self._thread = None

    def main(self) -> None:
        """Main worker thread

        This method defined the main application processing.
        The main thread must monitor and respond to the self._shutdown event.

        This method *must* be overridded by subclasses.
        """
        seconds = 0
        while not self._shutdown.wait(1.0):
            logger.info("Main thread processing: {}".format(seconds))
            seconds += 1

        logger.info("BaseApplication {} received shutdown event".format(self.name))
        self._shutdown.clear()



import time

import pytest

from telliot_core.directory import contract_directory
from telliot_core.directory import ContractInfo


def test_contract_info():
    """Test the ContractInfo class"""
    c = ContractInfo(
        org="tellor",
        name="tellorx-master",
        address={
            1: "0x88dF592F8eb5D7Bd38bFeF7dEb0fBc02cf3778a0",
            4: "0x88dF592F8eb5D7Bd38bFeF7dEb0fBc02cf3778a0",
        },
        abi_file="tellorx-master-abi.json",
    )

    # Make sure we can get ABI
    assert isinstance(c.get_abi(), list)


def test_contract_directory():
    """Test the contract directory"""
    cd = contract_directory
    assert len(cd.entries) > 6

    master = cd.find(name="tellorx-master")[0]
    assert isinstance(master.get_abi(), list)

    treasury = cd.find(address="0x2dB91443f2b562B8b2B2e8E4fC0A3EDD6c195147")[0]
    assert isinstance(treasury.get_abi(0), list)

    tellorx = cd.find(name="tellorx")
    assert len(tellorx) == 5

    mainnet_contracts = cd.find(chain_id=1)
    assert isinstance(mainnet_contracts[0], ContractInfo)


def test_directory_config_file():
    """Test the contract directory config file"""
    cd = contract_directory
    assert len(cd.entries) > 6


@pytest.mark.skip("Runs into API rate limits with other tests")
def test_abi_mainnet_retrieval():
    """Test the ABI retrieval from mainnet"""
    cd = contract_directory
    entries = cd.find(name="tellor-provider-id41", chain_id=1)
    info = entries[0]
    print(info)
    abi = info.get_abi(chain_id=1)
    assert abi[0]["name"] == "tellorReport"


@pytest.mark.skip("Long test")
def test_abis():
    """Make sure that ABI is provided or can be retrieved for all contracts.
    Very slow due to etherscan rate limits with no api_key.
    """
    for info in contract_directory.entries.values():
        print(f"Getting ABI for {info.name}")
        time.sleep(2)
        abi = info.get_abi()
        assert isinstance(abi, list)
        assert len(abi) > 0



"""
Unit tests covering telliot_core config options.
"""
import os
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path

from telliot_core.apps.config import ConfigFile
from telliot_core.apps.config import ConfigOptions


@dataclass
class SubConfig(ConfigOptions):
    option_c: str = "abc"


@dataclass
class MyOptions(ConfigOptions):
    option_a: int = 1
    option_b: SubConfig = field(default_factory=SubConfig)


def main(config_format):
    """Test default configs"""

    # Make sure that config file does not exist
    tmpdir = Path("./temp").resolve().absolute()
    config_file = tmpdir / f"myconfig.{config_format}"
    if config_file.exists():
        os.remove(config_file)
    for filename in config_file.parent.glob("myconfig*.bak"):
        filename.unlink()

    # Create folder if it doesn't exist
    tmpdir.mkdir(parents=True, exist_ok=True)

    # Make sure starting with clean folder
    assert not config_file.exists()
    for filename in config_file.parent.glob("myconfig*.bak"):
        filename.unlink()

    cf = ConfigFile(
        name="myconfig",
        config_type=MyOptions,
        config_dir=tmpdir,
        config_format=config_format,
    )

    # Make sure default was created and verify default value
    assert cf.config_file.exists()
    options = cf.get_config()
    assert options.option_a == 1

    # Make a change to the config and save it
    options.option_a = 3
    cf.save_config(options)

    # Create a new config object from the existing file.  Verify new values
    cf2 = ConfigFile(
        name="myconfig",
        config_type=MyOptions,
        config_dir=tmpdir,
        config_format=config_format,
    )
    op2 = cf2.get_config()
    assert op2.option_a == 3

    # Cleanup
    os.remove(config_file)
    for filename in config_file.parent.glob("myconfig*.bak"):
        filename.unlink()


def test_yaml():
    """Test YAML format"""
    main("yaml")


def test_json():
    """Test JSON format"""
    main("json")



import asyncio

import pytest

from telliot_core.apps.session_manager import ClientSessionManager


# ---------------------------------------------------------------------------------------
# BEGIN Asyncio windows bug workaround
#
# The first solution silences the exception that is raised on windows
#   https://github.com/aio-libs/aiohttp/issues/4324#issuecomment-733884349
# The second adds a delay before closing the event_loop (see event_loop below)
#   https://github.com/encode/httpx/issues/914#issuecomment-780023632
# Should only need to do one of these.
# ---------------------------------------------------------------------------------------
# def silence_event_loop_closed(func):
#     @wraps(func)
#     def wrapper(self, *args, **kwargs):
#         try:
#             return func(self, *args, **kwargs)
#         except RuntimeError as e:
#             if str(e) != 'Event loop is closed':
#                 raise
#
#     return wrapper
#
#
# import platform
# from asyncio.proactor_events import _ProactorBasePipeTransport
# if platform.system() == 'Windows':
#     _ProactorBasePipeTransport.__del__ =
#              silence_event_loop_closed(_ProactorBasePipeTransport.__del__)


@pytest.fixture(scope="module")
def event_loop():
    """Override event loop

    Reason 1: Override scope to match client_session scope
    Reason 2: Add delay before closing
            https://github.com/encode/httpx/issues/914#issuecomment-780023632
    """
    loop = asyncio.get_event_loop_policy().new_event_loop()
    yield loop
    loop.run_until_complete(asyncio.sleep(0.1))
    loop.close()


@pytest.fixture(scope="module")
async def client_session():
    """Reusable client session"""
    cm = ClientSessionManager()
    await cm.open()
    yield cm
    await asyncio.sleep(0)
    await cm.close()


#   E       AttributeError: 'async_generator' object has no attribute 'fetch_json'
@pytest.mark.skip("Does aiottp have support for async generators?")
@pytest.mark.asyncio
async def test_session(client_session):
    """Test the ClientSessionManager class"""
    result = await client_session.fetch_json(
        "https://api.coingecko.com/api/v3/simple/price?ids=bitcoin&vs_currencies=usd"
    )
    assert "bitcoin" in result



from dataclasses import dataclass
from typing import Callable
from typing import Optional

from telliot_core.model.base import Base


@dataclass
class ResponseStatus(Base):
    ok: bool = True
    error: Optional[str] = None
    e: Optional[Exception] = None


def error_status(
    note: str,
    e: Optional[Exception] = None,
    log: Optional[Callable[[str], None]] = None,
) -> ResponseStatus:
    """Helper for reporting and logging error status"""

    status = ResponseStatus()
    status.ok = False

    if e:
        msg = f"{note}: {repr(e)}"
    else:
        status.e = e
        msg = f"{note}"

    status.error = str(msg)

    if log:
        log(msg)

    return status



import os
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from typing import Literal
from typing import Optional
from typing import Union

from chained_accounts import ChainedAccount
from chained_accounts import find_accounts

from telliot_core.apps.config import ConfigFile
from telliot_core.apps.config import ConfigOptions
from telliot_core.model.api_keys import ApiKeyList
from telliot_core.model.base import Base
from telliot_core.model.chain import ChainList
from telliot_core.model.endpoints import EndpointList
from telliot_core.model.endpoints import RPCEndpoint


@dataclass
class MainConfig(ConfigOptions):
    """Main telliot_core configuration object"""

    #: Control application logging level
    loglevel: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"] = "INFO"

    #: Select chain id
    chain_id: int = 4


@dataclass
class TelliotConfig(Base):
    """Main telliot_core configuration object

    Load all configuration objects from config files.
    If any config file does not exist, a default will be created.
    """

    config_dir: Optional[Union[str, Path]] = None

    main: MainConfig = field(default_factory=MainConfig)

    endpoints: EndpointList = field(default_factory=EndpointList)

    api_keys: ApiKeyList = field(default_factory=ApiKeyList)

    chains: ChainList = field(default_factory=ChainList)

    # Private storage for config files
    _main_config_file: Optional[ConfigFile] = None
    _ep_config_file: Optional[ConfigFile] = None
    _api_keys_config_file: Optional[ConfigFile] = None
    _chain_config_file: Optional[ConfigFile] = None

    def __post_init__(self) -> None:
        self._main_config_file = ConfigFile(
            name="main",
            config_type=MainConfig,
            config_format="yaml",
            config_dir=self.config_dir,
        )
        self._ep_config_file = ConfigFile(
            name="endpoints",
            config_type=EndpointList,
            config_format="yaml",
            config_dir=self.config_dir,
        )
        self._api_keys_config_file = ConfigFile(
            name="api_keys",
            config_type=ApiKeyList,
            config_format="yaml",
            config_dir=self.config_dir,
        )
        self._chain_config_file = ConfigFile(
            name="chains",
            config_type=ChainList,
            config_format="json",
            config_dir=self.config_dir,
        )

        self.main = self._main_config_file.get_config()
        self.endpoints = self._ep_config_file.get_config()
        self.api_keys = self._api_keys_config_file.get_config()
        self.chains = self._chain_config_file.get_config()

    def get_endpoint(self) -> RPCEndpoint:
        """Search endpoints for current chain_id"""
        eps = self.endpoints.find(chain_id=self.main.chain_id)
        if len(eps) > 0:
            return eps[0]
        else:
            raise ValueError(f"Endpoint not found for chain_id={self.main.chain_id}")


def override_test_config(cfg: TelliotConfig, write: bool = False) -> TelliotConfig:
    """Override config with test configuration

    Returns a rinkeby test configuration, using github secrets if they are defined

    FOR DEVELOPMENT USE ONLY
    Overrides the current configuration with rinkeby test config.
    Also handles overrides for github secret keys.
    """

    # Override configuration for rinkeby testnet
    override_main = False
    if cfg.main.chain_id != 4:
        cfg.main.chain_id = 4
        override_main = True

    rinkeby_endpoint = cfg.get_endpoint()
    assert rinkeby_endpoint is not None

    override_endpoint = False
    if os.getenv("NODE_URL", None):
        rinkeby_endpoint.url = os.environ["NODE_URL"]
        override_endpoint = True

    rinkeby_accounts = find_accounts(chain_id=4)
    if not rinkeby_accounts:

        # Add private key if detected on git
        if os.getenv("PRIVATE_KEY", None):
            # Create an account for use on git
            ChainedAccount.add("git-rinkeby-key", chains=[4], key=os.environ["PRIVATE_KEY"], password="")

    if write:
        if override_endpoint:
            assert cfg._ep_config_file is not None
            cfg._ep_config_file.save_config(cfg.endpoints)
        if override_main:
            assert cfg._main_config_file is not None
            cfg._main_config_file.save_config(cfg.main)

    return cfg



import logging
from typing import Optional
from typing import Tuple

from chained_accounts import ChainedAccount

from telliot_core.contract.contract import Contract
from telliot_core.directory import contract_directory
from telliot_core.model.endpoints import RPCEndpoint
from telliot_core.utils.response import ResponseStatus
from telliot_core.utils.timestamp import TimeStamp


logger = logging.getLogger(__name__)


class Tellor360OracleContract(Contract):
    def __init__(self, node: RPCEndpoint, account: Optional[ChainedAccount] = None):
        chain_id = node.chain_id
        assert chain_id is not None

        contract_info = contract_directory.find(chain_id=chain_id, name="tellor360-oracle")
        if not contract_info:
            raise Exception(f"Tellor360 oracle contract not found on chain_id {chain_id}")

        contract_abi = contract_info[0].get_abi(chain_id=chain_id)

        super().__init__(
            address=contract_info[0].address[chain_id],
            abi=contract_abi,
            node=node,
            account=account,
        )

    async def get_time_of_last_new_value(self) -> Tuple[Optional[TimeStamp], ResponseStatus]:

        tlnv, status = await self.read("getTimeOfLastNewValue")

        if status.ok:
            return TimeStamp(tlnv), status
        else:
            return None, status



"""
Tests covering Pytelliot data source API key utils.
"""
from telliot_core.model.api_keys import ApiKeyList


def test_api_key_list():
    lis = ApiKeyList()
    # print(json.dumps(sl.get_state(), indent=2))
    res = lis.find(name="bravenewcoin")[0]
    assert res.url == "https://bravenewcoin.p.rapidapi.com/"

    res = lis.find(url="https://api.nomics.com/")[0]
    assert res.name == "nomics"

    res = lis.find(name="coinmarketcap")[0]
    assert res.url == "https://pro-api.coinmarketcap.com/"



import getpass

from chained_accounts import ChainedAccount
from hexbytes import HexBytes


def ask_for_password(name: str) -> str:
    password1 = getpass.getpass(f"Enter encryption password for {name}: ")
    password2 = getpass.getpass("Confirm password: ")
    if password2 != password1:
        raise Exception(f"Passwords do not match: {name}")
    password = password1

    return password


def lazy_unlock_account(account: ChainedAccount) -> None:
    if account.is_unlocked:
        return
    else:
        # Try unlocking with no password
        try:
            account.unlock("")
        except ValueError:
            try:
                password = getpass.getpass(f"Enter encryption password for {account.name}: ")
                account.unlock(password)
            except ValueError:
                raise Exception(f"Invalid password for {account.name}")


def lazy_key_getter(account: ChainedAccount) -> HexBytes:

    lazy_unlock_account(account)

    return account.key  # type: ignore



""" Listener Module

This module adds contract event listening capability for websocket endpoints
(like infura) that is lacking from web3.py.

The goal is to allow users to subscribe to and define async callbacks for each event.

*Work in Progress*
"""
import asyncio
import logging
import warnings
from typing import Any
from typing import Awaitable
from typing import Callable
from typing import List
from typing import Literal

import aiohttp
from aiohttp.client import _WSRequestContextManager
from hexbytes import HexBytes
from web3._utils.method_formatters import block_formatter
from web3._utils.method_formatters import log_entry_formatter
from web3._utils.method_formatters import syncing_formatter

logger = logging.getLogger(__name__)

AsyncCallable = Callable[[Any], Awaitable]

SubscriptionType = Literal["newHeads", "logs", "newPendingTransactions", "syncing"]


async def receive_message_task(
    ws: aiohttp.ClientWebSocketResponse,
    handler: AsyncCallable,
    formatter: Callable[..., Any],
) -> None:
    """A long running task that listens for JSON messages on an open websocket.
    The handler is called once for each message.

    Args:
        ws: Open websocket connection
        handler: An async function that takes the JSON message as a single argument
        formatter: An optional message formatter

    Returns: None
        Does not return until task is cancelled.

    """
    while True:
        try:
            message = await asyncio.wait_for(ws.receive_json(), timeout=60)
            # sub_id = message["params"]["subscription"]
            # logger.info(f"handled block subscription: {sub_id}")
            result = message["params"]["result"]
            formatted_result = formatter(result)
            asyncio.create_task(handler(formatted_result))  # type: ignore

        except asyncio.CancelledError:
            logger.debug("Listener cancelled")
            break
        except asyncio.exceptions.TimeoutError:
            continue


async def eth_subscribe(
    *,
    ws: aiohttp.ClientWebSocketResponse,
    name: SubscriptionType,
    lid: int = 1,
    **kwargs: Any,
) -> HexBytes:
    """Subscribe

    Args:
        ws: Websocket connection
        lid: Listener ID
        name: Subscription type/name, one of "newHeads", "logs", "newPendingTransactions", "syncing"
        **kwargs: Subscription parameter dict.  See (see https://geth.ethereum.org/docs/rpc/pubsub)

    Returns:
        Subscription ID

    """

    logger.debug(f"New {name} subscription")

    if kwargs:
        msg = {
            "jsonrpc": "2.0",
            "id": lid,
            "method": "eth_subscribe",
            "params": [name, kwargs],
        }
    else:
        msg = {"jsonrpc": "2.0", "id": lid, "method": "eth_subscribe", "params": [name]}
    await ws.send_json(msg)

    subscription_response = await ws.receive_json()

    lid_rx = subscription_response.get("id")
    assert lid_rx == lid

    sub_result = subscription_response.get("result")
    if not sub_result:
        logger.error("Subscription failed:")
        logger.error(subscription_response)
        raise Exception("Subscription Failed")

    return HexBytes(sub_result)


class Listener:
    def __init__(self, *, session: aiohttp.ClientSession, ws_url: str):

        self._session = session

        self._url: str = ws_url
        self._tasks: List[asyncio.Task[None]] = []
        self._listener_id: int = 0

    def connect(self) -> _WSRequestContextManager:  # aiohttp.ClientWebSocketResponse:
        return self._session.ws_connect(self._url)

    def _get_listener_id(self) -> int:
        """Generate sequential IDs for subscription requests"""
        self._listener_id += 1
        return self._listener_id

    async def eth_subscribe(
        self,
        handler: AsyncCallable,
        name: SubscriptionType,
        formatter: Callable[..., Any],
        **kwargs: Any,
    ) -> None:
        """Create a subscription using eth_subscribe

        Args:
            handler:
                An async function that takes the JSON message as a single argument
            name:
                Subscription type/name, one of:
                    "newHeads", "logs", "newPendingTransactions", "syncing"
            **kwargs:
                Subscription parameter dict.
                See (see https://geth.ethereum.org/docs/rpc/pubsub)

        Returns:

        """

        # Create a unique ID/name for the subscription
        lid = self._get_listener_id()
        task_name = f"listener_task_{lid}"

        # Launch a task and store a reference locally for later cancellation
        # Note that storing this reference delays exceptions until garbage collection.
        # Therefore, task is wrapped with a task_done callback to immediately catch exceptions
        task = asyncio.create_task(
            self._eth_subscribe_task(handler=handler, name=name, lid=lid, formatter=formatter, **kwargs),
            name=task_name,
        )

        task.add_done_callback(_handle_task_result)
        self._tasks.append(task)

    async def subscribe_new_blocks(self, handler: AsyncCallable) -> None:

        await self.eth_subscribe(handler=handler, name="newHeads", formatter=block_formatter)

    async def subscribe_contract_events(self, handler: AsyncCallable, address: str) -> None:

        await self.eth_subscribe(handler=handler, name="logs", address=address, formatter=log_entry_formatter)

    async def subscribe_pending_transactions(self, handler: AsyncCallable) -> None:

        await self.eth_subscribe(
            handler=handler,
            name="newPendingTransactions",
            formatter=pending_transaction_formatter,
        )
        # await self.eth_subscribe(handler=handler, name='newPendingTransactions')

    async def subscribe_syncing(self, handler: AsyncCallable) -> None:

        await self.eth_subscribe(handler=handler, name="syncing", formatter=syncing_formatter)

    async def _eth_subscribe_task(
        self,
        handler: AsyncCallable,
        name: SubscriptionType,
        lid: int,
        formatter: Callable[..., Any],
        **kwargs: Any,
    ) -> None:
        """A long running listener task.

        Note: Does not return until asyncio.Cancelled event
        """

        async with self.connect() as ws:
            sub_result = await eth_subscribe(ws=ws, name=name, lid=lid, **kwargs)
            if name == "logs":
                logger.info(f"Subscribed to contract address={kwargs['address']} (id={sub_result.hex()})")
            else:
                logger.info(f"New {name} subscription (id={sub_result.hex()})")
            await receive_message_task(ws=ws, handler=handler, formatter=formatter)

    async def shutdown(self) -> None:
        """Shut down all asyncio subscription tasks"""
        for task in self._tasks:
            logger.info(f"Shutting down listener {task.get_name()}")
            task.cancel()
        await asyncio.gather(*self._tasks)
        self._tasks = []

    def __del__(self) -> None:
        if self._tasks:
            warnings.warn("Listener.shutdown() not awaited.")


def _handle_task_result(task: asyncio.Task[Any]) -> None:
    # https://quantlane.com/blog/ensure-asyncio-task-exceptions-get-logged/
    try:
        task.result()
    except asyncio.CancelledError:
        pass  # Task cancellation should not be logged as an error.
    except Exception:
        logger.exception("Exception raised by task = %r", task)


async def event_logger(log: Any) -> None:
    # sub_id = msg["params"]["subscription"]
    # logger.info(f"handled block subscription: {sub_id}")
    # log_entry = msg["params"]["result"]
    # log_receipt: LogReceipt = log_entry_formatter(log_entry)
    logger.info(f"handled event message: {log}")


async def block_logger(block: Any) -> None:
    # sub_id = msg["params"]["subscription"]
    # logger.info(f"handled block subscription: {sub_id}")
    # block = msg["params"]["result"]
    logger.info(f"New block: {block}")


async def pending_transaction_logger(msg: Any) -> None:
    logger.info(f"New pending transaction: {msg!r}")


def pending_transaction_formatter(hash_str: str) -> HexBytes:
    return HexBytes(hash_str)


async def syncing_logger(msg: Any) -> None:
    logger.info(f"New sync: {msg}")


if __name__ == "__main__":
    from telliot_core.apps.core import TelliotCore
    import asyncio

    logging.basicConfig(level=logging.INFO)

    async def main() -> None:
        async with TelliotCore() as core:
            master_info = core.config.directory.find(name="tellorx-master", chain_id=core.config.main.chain_id)[0]
            oracle_info = core.config.directory.find(name="tellorx-oracle", chain_id=core.config.main.chain_id)[0]

            # Subscribe to blocks
            assert core.listener  # typing
            await core.listener.subscribe_new_blocks(handler=block_logger)
            # await core.listener.subscribe_syncing(handler=syncing_logger)

            # Subscribe to contract events
            await core.listener.subscribe_contract_events(
                handler=event_logger,
                address=master_info.address[core.config.main.chain_id],
            )
            await core.listener.subscribe_contract_events(
                handler=event_logger,
                address=oracle_info.address[core.config.main.chain_id],
            )

            # Subscribe to pending transactions:
            # Warning: Very high RPC transaction rate
            await core.listener.subscribe_pending_transactions(handler=pending_transaction_logger)

            await asyncio.sleep(1011)

    asyncio.run(main())



import logging
from typing import Any
from typing import Optional
from typing import Tuple

from chained_accounts import ChainedAccount

from telliot_core.contract.contract import Contract
from telliot_core.directory import contract_directory
from telliot_core.model.endpoints import RPCEndpoint
from telliot_core.utils.response import ResponseStatus
from telliot_core.utils.timestamp import TimeStamp


logger = logging.getLogger(__name__)


class TellorFlexOracleContract(Contract):
    def __init__(self, node: RPCEndpoint, account: Optional[ChainedAccount] = None):
        chain_id = node.chain_id
        assert chain_id is not None

        entries = contract_directory.find(chain_id=chain_id, name="tellorflex-oracle")
        if not entries:
            raise Exception(f"Tellorflex oracle contract not found on chain_id {chain_id}")
        contract_info = entries[0]

        contract_abi = contract_info.get_abi(chain_id=chain_id)

        super().__init__(
            address=contract_info.address[chain_id],
            abi=contract_abi,
            node=node,
            account=account,
        )

    async def get_governance_address(self) -> Optional[str]:

        governance_address, status = await self.read("getGovernanceAddress")

        if status.ok:
            return str(governance_address)
        else:
            logger.error("Error reading TellorFlexOracleContract")
            logger.error(status)
            return None

    async def get_reporting_lock(self) -> Optional[int]:

        lock, status = await self.read("getReportingLock")

        if status.ok:
            return int(lock)
        else:
            logger.error("Error reading TellorFlexOracleContract")
            logger.error(status)
            return None

    async def get_stake_amount(self) -> Optional[float]:

        stake, status = await self.read("getStakeAmount")

        if status.ok:
            stake_in_trb = int(stake) / 1.0e18
            return stake_in_trb
        else:
            logger.error("Error reading TellorFlexOracleContract")
            logger.error(status)
            return None

    async def get_time_of_last_new_value(self) -> Tuple[Optional[TimeStamp], ResponseStatus]:

        tlnv, status = await self.read("getTimeOfLastNewValue")

        if status.ok:
            return TimeStamp(tlnv), status
        else:
            return None, status

    async def get_token_address(self) -> Optional[str]:

        token_address, status = await self.read("getTokenAddress")

        if status.ok:
            return str(token_address)
        else:
            logger.error("Error reading TellorFlexOracleContract")
            logger.error(status)
            return None

    async def get_total_stake_amount(self) -> Optional[float]:

        total_stake, status = await self.read("getTotalStakeAmount")

        if status.ok:
            total_stake_trb = int(total_stake) / 1.0e18
            return total_stake_trb
        else:
            logger.error("Error reading TellorFlexOracleContract")
            logger.error(status)
            return None

    async def get_staker_info(self, staker_address: str) -> Tuple[Optional[Any], ResponseStatus]:

        staker_info, status = await self.read(func_name="getStakerInfo", _staker=staker_address)

        return staker_info, status

    async def get_new_value_count_by_qeury_id(self, query_id: bytes) -> Tuple[int, ResponseStatus]:
        count, status = await self.read(func_name="getNewValueCountbyQueryId", _queryId=query_id)
        return count, status


if __name__ == "__main__":
    import asyncio
    from telliot_core.apps.core import TelliotCore

    async def hello_world() -> None:
        async with TelliotCore(chain_id=137) as core:

            flex = core.get_tellorflex_contracts()

            t = await flex.oracle.get_time_of_last_new_value()

            print(f"Hello world!  Data was just submitted to TellorFlex on: {t}")

    asyncio.run(hello_world())



import asyncio
import logging

import pytest

from telliot_core.apps.core import TelliotCore
from telliot_core.contract.listener import block_logger


async def block_printer(msg) -> None:
    """Print block info"""
    sub_id = msg["params"]["subscription"]
    block = msg["params"]["result"]
    print(f"handled block subscription: {sub_id}")
    print(block)


@pytest.mark.asyncio
async def test_subscribe_new_blocks(caplog, mumbai_test_cfg):
    """Test the block listener"""
    caplog.set_level(logging.INFO)
    async with TelliotCore(config=mumbai_test_cfg) as core:
        # Subscribe to blocks
        await core.listener.subscribe_new_blocks(handler=block_logger)
        await asyncio.sleep(1)  # Note delay needed

    print(caplog.text)



import logging
import sys
from logging.handlers import RotatingFileHandler

from telliot_core.utils.home import default_homedir

# from telliot_core.plugin.discover import telliot_plugins

log_core = logging.getLogger("telliot_core")


def init_logging(level: int) -> logging.Logger:
    """Initialize logging to console and telliot logfile."""
    log_dir = default_homedir() / "logs"
    log_dir.mkdir(exist_ok=True)
    log_filename = log_dir / ("telliot.log")
    log_format = "%(levelname)-7s | %(name)s | %(message)s"
    formatter = logging.Formatter(log_format)

    log_core.setLevel(logging.DEBUG)

    # File Handler
    one_mb = 1024 * 1024
    fh = RotatingFileHandler(log_filename, maxBytes=one_mb * 5, backupCount=5)
    fh.setLevel(logging.DEBUG)
    fh.setFormatter(formatter)

    # Console handler
    stream = logging.StreamHandler(sys.stdout)
    stream.setLevel(level)
    stream.setFormatter(formatter)

    # Add handlers to core logger
    log_core.addHandler(fh)
    log_core.addHandler(stream)

    # # Add handlers to plugins
    # for plugin_name in telliot_plugins.keys():
    #     plugin_logger = logging.getLogger(plugin_name)
    #     plugin_logger.setLevel(logging.DEBUG)
    #     plugin_logger.addHandler(stream)
    #     plugin_logger.addHandler(fh)
    #     log_core.debug(f"Configured logging for {plugin_name} plugin")

    return log_core



import logging
from typing import Optional

from chained_accounts import ChainedAccount

from telliot_core.contract.contract import Contract
from telliot_core.directory import contract_directory
from telliot_core.model.endpoints import RPCEndpoint

logger = logging.getLogger(__name__)


class TokenContract(Contract):
    def __init__(self, node: RPCEndpoint, account: Optional[ChainedAccount] = None):
        chain_id = node.chain_id
        assert chain_id is not None

        if chain_id == 122:
            contract_info = contract_directory.find(chain_id=chain_id, name="wrapped-fuse-token")[0]
        else:
            contract_info = contract_directory.find(chain_id=chain_id, name="trb-token")[0]

        if not contract_info:
            raise Exception(f"Tellorflex token contract not found on chain_id {chain_id}")

        contract_abi = contract_info.get_abi(chain_id=chain_id)

        super().__init__(
            address=contract_info.address[chain_id],
            abi=contract_abi,
            node=node,
            account=account,
        )



import asyncio
from functools import wraps

import click

from telliot_core.apps.core import TelliotCore
from telliot_core.apps.telliot_config import override_test_config
from telliot_core.apps.telliot_config import TelliotConfig


def async_run(f):  # type: ignore
    """Call and run an async function.

    Handy Click CLI tests of async functions."""

    @wraps(f)
    def wrapper(*args, **kwargs):  # type: ignore
        return asyncio.run(f(*args, **kwargs))

    return wrapper


def cli_config(ctx: click.Context) -> TelliotConfig:
    """Return a telliot configuration using the CLI context"""
    if ctx.obj["TEST_CONFIG"]:
        cfg = override_test_config(TelliotConfig())

    else:
        cfg = TelliotConfig()
        if ctx.obj.get("CHAIN_ID", None):
            cfg.main.chain_id = ctx.obj["CHAIN_ID"]

    return cfg


def cli_core(ctx: click.Context) -> TelliotCore:
    """Returns a TelliotCore configured with the CLI context

    The returned object should be used as a context manager for CLI commands
    """
    account_name = ctx.obj.get("ACCOUNT_NAME", None)

    cfg = cli_config(ctx)
    core = TelliotCore(config=cfg, account_name=account_name)
    return core



""" Telliot CLI

A simple interface for interacting with telliot_core's functionality.
Configure telliot_core's settings via this interface's command line flags
or in the configuration file.
"""
import click

from telliot_core.cli.commands.account import account
from telliot_core.cli.commands.config import config
from telliot_core.cli.commands.listen import listen
from telliot_core.cli.commands.read import read
from telliot_core.utils.versions import show_telliot_versions


@click.group(invoke_without_command=True)
@click.pass_context
@click.option(
    "--chain_id",
    type=int,
    help="Override chain ID (the default is provided by main config file).",
)
@click.option(
    "--test_config",
    is_flag=True,
    help="Runs command with test configuration (developer use only)",
)
@click.option("--version", is_flag=True, help="Display telliot-core version and exit.")
def main(ctx: click.Context, version: bool, chain_id: int, test_config: bool) -> None:
    ctx.ensure_object(dict)
    ctx.obj["CHAIN_ID"] = chain_id
    ctx.obj["TEST_CONFIG"] = test_config
    if version:
        show_telliot_versions()
        return

    """Telliot command line interface"""
    if ctx.invoked_subcommand is None:
        print(ctx.command.get_help(ctx))


main.add_command(config)
main.add_command(read)
main.add_command(account)
main.add_command(listen)

if __name__ == "__main__":
    main()



import pytest
from brownie import accounts
from brownie import TellorXOracleMock

from telliot_core.apps.core import TelliotCore
from telliot_core.tellor.tellorx.oracle import TellorxOracleContract
from telliot_core.utils.timestamp import TimeStamp


@pytest.fixture
def tellorx_oracle_mock_contract():
    """Reusable fixture for TellorXOracleMock contract"""
    return accounts[0].deploy(TellorXOracleMock)


@pytest.mark.asyncio
async def test_getReportTimestampByIndex(rinkeby_test_cfg, tellorx_oracle_mock_contract):
    """Test the TellorxOracleContract getReportTimestampByIndex method"""
    async with TelliotCore(config=rinkeby_test_cfg) as core:
        account = core.get_account()
        oracle = TellorxOracleContract(core.endpoint, account)
        oracle.address = tellorx_oracle_mock_contract.address  # Override with locally-deployed mock contract address
        oracle.connect()

        queryId = "0x0000000000000000000000000000000000000000000000000000000000000001"
        index = 2
        timestamp, status = await oracle.getReportTimestampByIndex(queryId, index)
        assert status.ok
        assert isinstance(timestamp, int)
        print(timestamp)
        assert timestamp == 1234


@pytest.mark.asyncio
async def test_getReportingLock(rinkeby_test_cfg, tellorx_oracle_mock_contract):
    """Test the TellorxOracleContract.getReportingLock() method"""
    async with TelliotCore(config=rinkeby_test_cfg) as core:
        account = core.get_account()
        oracle = TellorxOracleContract(core.endpoint, account)
        oracle.address = tellorx_oracle_mock_contract.address  # Override with locally-deployed mock contract address
        oracle.connect()

        result, status = await oracle.getReportingLock()
        assert status.ok
        assert result == 12
        print(result)


@pytest.mark.asyncio
async def test_gettimebasedreward(rinkeby_test_cfg, tellorx_oracle_mock_contract):
    """Test the TellorxOracleContract.getTimeBasedReward() method"""
    async with TelliotCore(config=rinkeby_test_cfg) as core:
        account = core.get_account()
        oracle = TellorxOracleContract(core.endpoint, account)
        oracle.address = tellorx_oracle_mock_contract.address  # Override with locally-deployed mock contract address
        oracle.connect()

        result, status = await oracle.getTimeBasedReward()
        assert status.ok
        assert result == 1


@pytest.mark.asyncio
async def test_getCurrentReward(rinkeby_test_cfg, tellorx_oracle_mock_contract):
    """Test the TellorX oracle contract"""
    async with TelliotCore(config=rinkeby_test_cfg) as core:
        account = core.get_account()
        oracle = TellorxOracleContract(core.endpoint, account)
        oracle.address = tellorx_oracle_mock_contract.address  # Override with locally-deployed mock contract address
        oracle.connect()

        queryId = "0x0000000000000000000000000000000000000000000000000000000000000001"
        (tips, reward), status = await oracle.getCurrentReward(queryId)

        assert status.ok
        assert tips == 1
        assert reward == 2


@pytest.mark.asyncio
async def test_getTimestampCountById(rinkeby_test_cfg, tellorx_oracle_mock_contract):
    """Test the TellorX oracle contract"""
    async with TelliotCore(config=rinkeby_test_cfg) as core:
        account = core.get_account()
        oracle = TellorxOracleContract(core.endpoint, account)
        oracle.address = tellorx_oracle_mock_contract.address  # Override with locally-deployed mock contract address
        oracle.connect()

        queryId = "0x0000000000000000000000000000000000000000000000000000000000000001"
        result, status = await oracle.getTimestampCountById(queryId)

        assert status.ok
        assert result == 30


@pytest.mark.asyncio
async def test_getTimeOfLastNewValue(rinkeby_test_cfg, tellorx_oracle_mock_contract):
    """Test the TellorX oracle contract"""
    async with TelliotCore(config=rinkeby_test_cfg) as core:
        account = core.get_account()
        oracle = TellorxOracleContract(core.endpoint, account)
        oracle.address = tellorx_oracle_mock_contract.address  # Override with locally-deployed mock contract address
        oracle.connect()

        result, status = await oracle.getTimeOfLastNewValue()

        assert status.ok
        assert isinstance(result, TimeStamp)
        assert result.ts == 123456789


@pytest.mark.asyncio
async def test_getTipsById(rinkeby_test_cfg, tellorx_oracle_mock_contract):
    """Test the TellorX oracle contract"""
    async with TelliotCore(config=rinkeby_test_cfg) as core:
        account = core.get_account()
        oracle = TellorxOracleContract(core.endpoint, account)
        oracle.address = tellorx_oracle_mock_contract.address  # Override with locally-deployed mock contract address
        oracle.connect()

        queryId = "0x0000000000000000000000000000000000000000000000000000000000000002"
        result, status = await oracle.getTipsById(queryId)

        assert status.ok
        assert result == 3



import logging
from typing import Optional
from typing import Tuple

from chained_accounts import ChainedAccount
from web3.exceptions import ContractLogicError

from telliot_core.contract.contract import Contract
from telliot_core.directory import contract_directory
from telliot_core.model.endpoints import RPCEndpoint
from telliot_core.utils.response import ResponseStatus


logger = logging.getLogger(__name__)


class Tellor360AutopayContract(Contract):
    """Tellor360 Autopay contract getter"""

    def __init__(self, node: RPCEndpoint, account: Optional[ChainedAccount] = None):
        chain_id = node.chain_id
        assert chain_id is not None

        contract_info = contract_directory.find(chain_id=chain_id, name="tellor360-autopay")
        if not contract_info:
            raise Exception(f"Tellor360 autopay contract not found on chain_id {chain_id}")
        contract_abi = contract_info[0].get_abi(chain_id=chain_id)

        super().__init__(
            address=contract_info[0].address[chain_id],
            abi=contract_abi,
            node=node,
            account=account,
        )

    async def get_current_tip(self, query_id: bytes) -> Tuple[Optional[int], ResponseStatus]:
        tip_amount, status = await self.read(func_name="getCurrentTip", _queryId=query_id)
        if status.ok:
            return tip_amount, status
        # autopay contract reverts when tip amount is zero
        # instead of returning 0, not sure why
        elif type(status.e) == ContractLogicError:
            tip_amount = 0
            status.ok = True
            return tip_amount, status
        else:
            return None, status



import pytest
from brownie import accounts
from brownie import TellorXMasterMock

from telliot_core.apps.core import TelliotCore
from telliot_core.tellor.tellorx.master import account_status_map
from telliot_core.tellor.tellorx.master import TellorxMasterContract
from telliot_core.utils.timestamp import TimeStamp


@pytest.fixture
def tellorx_master_mock_contract():
    """Mock the TellorXMaster contract"""
    return accounts[0].deploy(TellorXMasterMock)


@pytest.mark.asyncio
async def test_get_staker_info(rinkeby_test_cfg, tellorx_master_mock_contract):
    """Test the TellorXMaster contract"""

    async with TelliotCore(config=rinkeby_test_cfg) as core:
        account = core.get_account()
        tellorx = TellorxMasterContract(core.endpoint, account)
        tellorx.address = tellorx_master_mock_contract.address
        tellorx.connect()

        result, status = await tellorx.getStakerInfo(tellorx.address)

        assert status.ok
        assert len(result) == 2
        assert result[0] in account_status_map.values()
        assert isinstance(result[1], TimeStamp)



"""
Config file for user API keys for data sources.
"""
import logging
from dataclasses import dataclass
from dataclasses import field
from typing import List
from typing import Optional

from telliot_core.apps.config import ConfigFile
from telliot_core.apps.config import ConfigOptions
from telliot_core.model.base import Base

logger = logging.getLogger(__name__)


@dataclass
class ApiKey(Base):
    """API Key used for a data source."""

    #: API key name
    name: str = ""

    #: API key (secret)
    key: str = ""

    #: Associated URL
    url: str = ""


default_api_keys = [
    ApiKey(name="anyblock", key="", url="https://api.anyblock.tools/"),
    ApiKey(
        name="bravenewcoin",
        key="",
        url="https://bravenewcoin.p.rapidapi.com/",
    ),
    ApiKey(name="nomics", key="", url="https://api.nomics.com/"),
    ApiKey(name="coinmarketcap", key="", url="https://pro-api.coinmarketcap.com/"),
]


@dataclass
class ApiKeyList(ConfigOptions):
    api_keys: List[ApiKey] = field(default_factory=lambda: default_api_keys)

    def find(
        self,
        name: Optional[str] = None,
        url: Optional[str] = None,
    ) -> list[ApiKey]:

        result = []
        for api_key in self.api_keys:

            if name is not None:
                if name != api_key.name:
                    continue
            if url is not None:
                if url != api_key.url:
                    continue

            result.append(api_key)

        return result


if __name__ == "__main__":
    cf = ConfigFile(name="api_keys", config_type=ApiKeyList, config_format="yaml")

    config_endpoints = cf.get_config()

    print(config_endpoints)



from typing import Optional

import click
import yaml

from telliot_core.apps.telliot_config import override_test_config
from telliot_core.apps.telliot_config import TelliotConfig


@click.group()
def config() -> None:
    """Manage Telliot configuration."""
    pass


@config.command()
def init() -> None:
    """Create initial configuration files."""
    _ = TelliotConfig()


@config.command()
def show() -> None:
    """Show current configuration."""
    cfg = TelliotConfig()
    state = cfg.get_state()

    print(yaml.dump(state, sort_keys=False))


@config.command()
@click.option("-c", "--chain_id", type=int, default=None, help="Chain ID")
def set(chain_id: Optional[int]) -> None:
    """Set a configuration parameter."""
    cfg = TelliotConfig()
    modified = False

    if chain_id:
        cfg.main.chain_id = chain_id
        modified = True

    if modified:
        assert cfg._main_config_file is not None
        cfg._main_config_file.save_config(cfg.main)


@config.command()
def testconfig() -> None:
    """Custom configuration for rinkeby testing on git.

    Modifies the configuration files to use a rinkeby test
    configuration, using github secrets if they are defined

    FOR DEVELOPMENT USE ONLY.
    """
    _ = override_test_config(TelliotConfig(), write=True)



from dataclasses import dataclass
from typing import Optional

from chained_accounts import ChainedAccount
from eth_utils import to_checksum_address

from telliot_core.contract.contract import Contract
from telliot_core.directory import contract_directory as directory
from telliot_core.model.endpoints import RPCEndpoint
from telliot_core.tellor.tellorx.oracle import ReadRespType
from telliot_core.utils.timestamp import TimeStamp

account_status_map = {
    0: "NotStaked",
    1: "Staked",
    2: "LockedForWithdraw",
    3: "InDispute",
    4: "disbursed",
    5: "slashed",
}


@dataclass
class DisputeReport:
    hash: str
    tally: int
    executed: bool
    disputeVotePassed: bool
    isPropFork: bool
    reportedMiner: str
    reportingParty: str
    proposedForkAddress: str


class TellorxMasterContract(Contract):
    def __init__(self, node: RPCEndpoint, account: Optional[ChainedAccount] = None):

        chain_id = node.chain_id
        assert chain_id is not None

        entries = directory.find(name="tellorx-master", chain_id=chain_id)
        if not entries:
            raise Exception(f"TellorX master contract not found on chain_id {chain_id}")
        contract_info = entries[0]
        contract_abi = contract_info.get_abi(chain_id=chain_id)

        super().__init__(
            address=contract_info.address[chain_id],
            abi=contract_abi,
            node=node,
            account=account,
        )

    async def getStakerInfo(self, address: str) -> ReadRespType:
        """Get Staker Info"""

        address = to_checksum_address(address)

        result, status = await self.read("getStakerInfo", _staker=address)

        if status.ok:
            current_status, ts_staked = result
            staker_status = account_status_map[current_status]
            date_staked = TimeStamp(ts_staked)
            return (staker_status, date_staked), status
        else:
            return (None, None), status

    async def disputesById(self, dispute_id: int) -> ReadRespType:

        response, status = await self.read("disputesById", dispute_id)

        if status.ok:
            result = DisputeReport(
                hash=f"0x{response[0].hex()}",
                tally=response[1],
                executed=response[2],
                disputeVotePassed=response[3],
                isPropFork=response[4],
                reportedMiner=response[5],
                reportingParty=response[6],
                proposedForkAddress=response[7],
            )
            return result, status
        else:
            return None, status



import pathlib
from pathlib import Path
from typing import Optional
from typing import Union

TELLIOT_CORE_ROOT = Path(__file__).parent.parent


def default_homedir() -> pathlib.Path:
    """Return default home directory, creating it if necessary

    Returns:
        pathlib.Path : Path to home directory
    """
    homedir = Path.home() / ("telliot")
    homedir = homedir.resolve().absolute()
    if not homedir.is_dir():
        homedir.mkdir()

    return homedir


def telliot_homedir(homedir: Optional[Union[str, Path]] = None) -> Path:
    """Telliot home directory

    Returns the telliot_core home Path, using a default if none is provided.
    The default directory is created if it does not exist.
    A homedir that is provided must exist
    """
    if homedir is None:
        # Set default homedir and create if necessary
        result = default_homedir()
        if not result.exists():
            result.mkdir(parents=True)
    else:
        # Use specified home directory
        if isinstance(homedir, str):
            homedir = Path(homedir)
        result = homedir.resolve().absolute()
        if not result.exists():
            raise FileExistsError("Directory does not exist: {}".format(homedir))

    return result



import json
from dataclasses import dataclass
from pathlib import Path
from typing import Dict
from typing import Optional

from clamfig import deserialize

from telliot_core.model.base import Base
from telliot_core.model.tokens import BlockChainAsset
from telliot_core.utils.home import TELLIOT_CORE_ROOT


@dataclass
class AssetRegistry(Base):
    assets: Dict[str, BlockChainAsset]

    def register(self, asset: BlockChainAsset) -> None:
        if asset.id in self.assets:
            raise ValueError(f"Asset {asset.id} already registered")
        self.assets[asset.id] = asset

    @classmethod
    def from_file(cls, filepath: Path) -> "AssetRegistry":

        with open(filepath) as f:
            state = json.load(f)

        asset_list = deserialize(state)

        obj = cls(assets={})

        for asset in asset_list:
            obj.register(asset)

        return obj

    def get(self, asset_id: str) -> Optional[BlockChainAsset]:

        return self.assets.get(asset_id)


asset_registry = AssetRegistry.from_file(TELLIOT_CORE_ROOT / "data/assets.json")



"""
Unit tests covering telliot_core CLI commands.

Note: Pytest Live Logs breaks some CliRunner tests.
The bug prevents CliRunner from capturing logger output
  - https://github.com/pallets/click/issues/2156
  - https://github.com/pallets/click/issues/824
"""
import pytest
from click.testing import CliRunner

from telliot_core.cli.main import main


def test_main_help():
    """Test telliot_core CLI command: report."""
    runner = CliRunner()
    result = runner.invoke(main, [])

    assert "Usage:" in result.stdout


def test_config_cmd():
    """Test telliot_core CLI command: report."""
    runner = CliRunner()
    result = runner.invoke(main, ["config", "init"])
    assert not result.exception
    print(result.stdout)


@pytest.mark.skip("Infura rinkeby node deprecated, needs to use brownie")
def test_disputesbyid(caplog):
    """Test disputes by ID"""
    runner = CliRunner()
    result = runner.invoke(main, ["--test_config", "read", "master", "disputesbyid", "1"])
    assert "DisputeReport" in result.stdout
    assert not result.exception


def test_getStakerInfo():
    """Test telliot_core CLI command: report."""
    runner = CliRunner()
    result = runner.invoke(main, ["--test_config", "read", "master", "getstakerinfo"])

    print(result.stdout)
    # expect a tuple of integers
    # output = eval(result.output.strip())
    # assert len(output) == 2
    # assert isinstance(output[0], int)
    # assert isinstance(output[1], int)


@pytest.mark.skip()
def test_gettimebasedreward():
    """Test contract method"""
    runner = CliRunner()
    result = runner.invoke(main, ["--test_config", "read", "gettimebasedreward"])
    assert "TRB" in result.output


def test_config_show():
    """Make sure config is running"""
    runner = CliRunner()
    result = runner.invoke(main, ["--test_config", "config", "show"])
    assert not result.exception


@pytest.mark.skip("Infura rinkeby node deprecated, needs to use brownie")
def test_account_status():
    """Test account status."""
    runner = CliRunner()
    result = runner.invoke(main, ["--test_config", "account", "status"])
    assert not result.exception



from telliot_core.model.tokens import ERC20TokenList


def test_token_list():
    """Test importing a file from uniswap format"""
    _ = ERC20TokenList.from_uniswap(EXAMPLE_TOKEN_LIST)


EXAMPLE_TOKEN_LIST = {
    "name": "My Token List",
    "version": {"major": 2, "minor": 0, "patch": 0},
    "tokens": [
        {
            "chainId": 1,
            "address": "0x7Fc66500c84A76Ad7e9c93437bFc5Ac33E2DDaE9",
            "name": "Aave",
            "symbol": "AAVE",
            "decimals": 18,
            "logoURI": "https://assets.coingecko.com/coins/images/12645/thumb/AAVE.png?1601374110",
        },
        {
            "chainId": 1,
            "address": "0xfF20817765cB7f73d4bde2e66e067E58D11095C2",
            "name": "Amp",
            "symbol": "AMP",
            "decimals": 18,
            "logoURI": "https://assets.coingecko.com/coins/images/12409/thumb/amp-200x200.png?1599625397",
        },
        {
            "name": "Aragon Network Token",
            "address": "0x960b236A07cf122663c4303350609A66A7B288C0",
            "symbol": "ANT",
            "decimals": 18,
            "chainId": 1,
            "logoURI": "https://raw.githubusercontent.com/trustwallet/"
            "assets/master/blockchains/ethereum/assets/"
            "0x960b236A07cf122663c4303350609A66A7B288C0/logo.png",
        },
    ],
}



import asyncio
from typing import Any
from typing import Optional

import aiohttp


class ClientSessionManager:
    """Session Manager

    Manage a single `aiohttp.ClientSession` for efficient
    handling of user requests for http access.
    """

    @property
    def session(self) -> aiohttp.ClientSession:
        """Returns the client session"""
        if self._session:
            return self._session
        else:
            raise Exception("Client session does not exist.  Use ClientSessionManager.open().")

    _session: Optional[aiohttp.ClientSession]

    def __init__(self) -> None:
        self._session = None

    async def open(self) -> None:
        """Create the client session."""
        self._session = aiohttp.ClientSession()

    async def close(self) -> None:
        """Close the client session."""
        if self.session:
            await self.session.close()

    async def fetch_json(self, url: str) -> Any:
        """Fetch JSON response from URL"""
        async with self.session.get(url) as resp:
            if resp.status == 200:
                json_obj = await resp.json()
                return json_obj
            else:
                return await None

    def __del__(self) -> None:
        """Make sure the client session is closed when this object is deleted."""
        if self._session:
            if not self._session.closed:
                try:
                    loop = asyncio.get_event_loop()
                    if loop.is_running():
                        loop.create_task(self.close())
                    else:
                        loop.run_until_complete(self.close())
                except Exception:
                    pass



from typing import Any
from typing import Dict


def dict2argstr(d: Dict[str, Any]) -> str:
    """Convert a dict to a text of kwd=arg pairs"""
    return ",".join("{!s}={!r}".format(key, val) for (key, val) in d.items())



from typing import Any
from typing import Callable

import telliot_core

# from telliot_core.plugin.discover import telliot_plugins


def show_telliot_versions(caller: Callable[[Any], None] = print) -> None:  # , include_plugins: bool = True) -> None:
    caller(f"telliot-core {telliot_core.__version__}")
    # if include_plugins:
    #     for name, pkg in telliot_plugins.items():
    #         if name != "telliot_core":
    #             try:
    #                 caller(f"{name} (plugin): Version {pkg.__version__}")
    #             except AttributeError:
    #                 caller(f"{name} (plugin): Version UNKNOWN")



"""
Utils for creating a JSON RPC connection to an EVM blockchain
"""
import logging
from dataclasses import dataclass
from dataclasses import field
from typing import List
from typing import Optional

import websockets.exceptions
from web3 import Web3
from web3.middleware import geth_poa_middleware

from telliot_core.apps.config import ConfigFile
from telliot_core.apps.config import ConfigOptions
from telliot_core.model.base import Base

logger = logging.getLogger(__name__)


@dataclass
class RPCEndpoint(Base):
    """JSON RPC Endpoint for EVM compatible network"""

    #: Chain ID
    chain_id: Optional[int] = None

    #: Network Name (e.g. 'mainnet', 'testnet', 'rinkeby')
    network: str = ""

    #: Provider Name (e.g. 'Infura')
    provider: str = ""

    #: URL (e.g. 'https://mainnet.infura.io/v3/<project_id>')
    url: str = ""

    #: Explorer URL ')
    explorer: Optional[str] = None

    #: Read-only Web3 Connection with private storage
    web3 = property(lambda self: self._web3)
    _web3: Optional[Web3] = field(default=None, init=False, repr=False)

    def connect(self) -> bool:
        """Connect to EVM blockchain

        A connection failure does not raise an exception.  This is left
        to the caller.

        returns:
            True if connection was successful
        """

        if self._web3:
            return True

        if self.url.startswith("ws"):
            self._web3 = Web3(Web3.WebsocketProvider(self.url))
        elif self.url.startswith("http"):
            self._web3 = Web3(Web3.HTTPProvider(self.url))
        else:
            raise ValueError(f"Invalid endpoint url: {self.url}")

        # Inject middleware if connecting to rinkeby (chain_id=4)
        if self.chain_id == 4:
            self.web3.middleware_onion.inject(geth_poa_middleware, layer=0)

        connected = False
        try:
            connected = self._web3.eth.get_block_number() > 1
            logger.debug("Connected to {}".format(self))

        except websockets.exceptions.InvalidStatusCode as e:
            connected = False
            msg = f"Could not connect to RPC endpoint at: {self.url}"
            logger.error(e)
            logger.error(msg)

        return connected


default_endpoint_list = [
    RPCEndpoint(
        chain_id=1,
        provider="Infura",
        network="mainnet",
        url="wss://mainnet.infura.io/ws/v3/{INFURA_API_KEY}",
        explorer="https://etherscan.io",
    ),
    RPCEndpoint(
        chain_id=3,
        provider="Infura",
        network="ropsten",
        url="wss://ropsten.infura.io/ws/v3/{INFURA_API_KEY}",
        explorer="https://ropsten.etherscan.io",
    ),
    RPCEndpoint(
        chain_id=4,
        provider="Infura",
        network="rinkeby",
        url="wss://rinkeby.infura.io/ws/v3/{INFURA_API_KEY}",
        explorer="https://rinkeby.etherscan.io",
    ),
    RPCEndpoint(
        chain_id=137,
        provider="Matic",
        network="mainnet",
        url="https://rpc-mainnet.matic.network",
        explorer="https://matic.network",
    ),
    RPCEndpoint(
        chain_id=122,
        provider="Fuse",
        network="mainnet",
        url="https://rpc.fuse.io",
        explorer="https://explorer.fuse.io",
    ),
    RPCEndpoint(
        chain_id=80001,
        provider="Matic",
        network="mumbai",
        url="https://rpc-mumbai.maticvigil.com",
        explorer="https://mumbai.polygonscan.com/",
    ),
    RPCEndpoint(
        chain_id=69,
        provider="optimism-kovan",
        network="infura",
        url="https://optimism-kovan.infura.io/v3/{INFURA_API_KEY}",
        explorer="https://kovan-optimistic.etherscan.io",
    ),
    RPCEndpoint(
        chain_id=1666600000,
        provider="Harmony",
        network="Harmony",
        url="https://api.harmony.one",
        explorer="https://explorer.harmony.one/",
    ),
    RPCEndpoint(
        chain_id=1666700000,
        provider="Harmony",
        network="Harmony Testnet",
        url="https://api.s0.b.hmny.io",
        explorer="https://explorer.pops.one/",
    ),
    RPCEndpoint(
        chain_id=421611,
        provider="Infura",
        network="Arbitrum Rinkeby",
        url="https://arbitrum-rinkeby.infura.io/v3/{INFURA_API_KEY}",
        explorer="https://rinkeby-explorer.arbitrum.io/#/",
    ),
    RPCEndpoint(
        chain_id=941,
        provider="Pulsechain",
        network="Pulsechain Testnet",
        url="https://rpc.v2b.testnet.pulsechain.com",
        explorer="https://scan.v2b.testnet.pulsechain.com",
    ),
    RPCEndpoint(
        chain_id=42161,
        provider="Infura",
        network="Arbitrum One",
        url="https://arbitrum-mainnet.infura.io/v3/{INFURA_API_KEY}",
        explorer="https://arbiscan.io",
    ),
]


@dataclass
class EndpointList(ConfigOptions):
    endpoints: List[RPCEndpoint] = field(default_factory=lambda: default_endpoint_list)

    def get_chain_endpoint(self, chain_id: int = 1) -> Optional[RPCEndpoint]:
        """Get an Endpoint for the specified chain_id"""

        for endpoint in self.endpoints:
            if endpoint.chain_id == chain_id:
                return endpoint

        return None

    def find(
        self,
        *,
        chain_id: Optional[int] = None,
        provider: Optional[str] = None,
    ) -> list[RPCEndpoint]:

        result = []
        for ep in self.endpoints:

            if chain_id is not None:
                if chain_id != ep.chain_id:
                    continue
            if provider is not None:
                if provider != ep.provider:
                    continue

            result.append(ep)

        return result


if __name__ == "__main__":
    cf = ConfigFile(name="endpoints", config_type=EndpointList, config_format="yaml")

    config_endpoints = cf.get_config()

    print(config_endpoints)



import logging
import pathlib
import shutil
from pathlib import Path

import pytest

from telliot_core.apps.core import TelliotCore
from telliot_core.utils.home import default_homedir

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)

# Temp file location
testhome = Path("./temp").resolve().absolute()


def test_homedir():
    """Test default home directory"""
    hd = default_homedir()
    assert isinstance(hd, pathlib.Path)
    assert hd.exists()


@pytest.mark.asyncio
async def test_application_homedir(mumbai_test_cfg):
    """Test home directory handling"""
    if testhome.exists():
        shutil.rmtree(testhome)

    testhome.mkdir(parents=True)
    app = TelliotCore(config=mumbai_test_cfg, homedir=testhome)
    assert isinstance(app.homedir, pathlib.Path)

    shutil.rmtree(testhome)


@pytest.mark.asyncio
async def test_application_default_home(mumbai_test_cfg):
    """Test default application directory"""

    async with TelliotCore(config=mumbai_test_cfg) as app:
        assert app.homedir == default_homedir()
        assert "telliot" in str(app.homedir)


@pytest.mark.asyncio
async def test_app_constructor(mumbai_test_cfg):
    tmpdir = Path(".tmp")
    if not tmpdir.exists():
        tmpdir.mkdir()

    # Create a default application
    async with TelliotCore(config=mumbai_test_cfg, homedir=Path(".tmp")) as app:
        assert app.homedir.absolute() == tmpdir.absolute()



import json
import logging
from json.decoder import JSONDecodeError
from typing import Literal
from typing import Optional

import requests

logger = logging.getLogger(__name__)
ethgastypes = Literal["fast", "fastest", "safeLow", "average"]


async def fetch_gas_price() -> Optional[int]:
    """Estimate current ETH gas price

    Current implementation fetches from ethgasstation

    Returns:
        eth gas price in gwei
    """
    return await ethgasstation("fast")


async def ethgasstation(style: ethgastypes = "fast", retries: int = 2) -> Optional[int]:
    """Fetch gas price from ethgasstation in gwei"""
    for _ in range(retries):
        try:
            rsp = requests.get("https://ethgasstation.info/json/ethgasAPI.json")
            prices = json.loads(rsp.content)
            gas_price = int(prices[style])
            return int(gas_price / 10)  # json output is gwei*10
        except JSONDecodeError:
            logger.error("Error decoding JSON from ethgasstation API")
            continue
        except requests.exceptions.SSLError:
            logger.error("SSLError -- Unable to fetch gas price")
            return None
        except Exception as e:
            logger.error(f"Error fetching gas price: {e}")
    return None



from typing import Any
from typing import Optional
from typing import Tuple

from chained_accounts import ChainedAccount

from telliot_core.contract.contract import Contract
from telliot_core.directory import contract_directory
from telliot_core.model.endpoints import RPCEndpoint
from telliot_core.utils.response import ResponseStatus
from telliot_core.utils.timestamp import TimeStamp

ReadRespType = Tuple[Any, ResponseStatus]


class TellorxOracleContract(Contract):
    def __init__(self, node: RPCEndpoint, account: Optional[ChainedAccount] = None):

        chain_id = node.chain_id
        assert chain_id is not None

        contract_info = contract_directory.find(chain_id=chain_id, name="tellorx-oracle")[0]
        if not contract_info:
            raise Exception(f"TellorX oracle contract not found on chain_id {chain_id}")

        contract_abi = contract_info.get_abi(chain_id=chain_id)

        super().__init__(
            address=contract_info.address[chain_id],
            abi=contract_abi,
            node=node,
            account=account,
        )

    async def getBlockNumberByTimestamp(self, queryId: str, timestamp: int) -> ReadRespType:

        result, status = await self.read("getBlockNumberByTimestamp", _queryId=queryId, _timestamp=timestamp)

        return result, status

    async def getCurrentReward(self, queryId: str) -> ReadRespType:

        result, status = await self.read("getCurrentReward", _queryId=queryId)

        if status.ok:
            (tips, reward) = result
            return (tips / 1.0e18, reward / 1.0e18), status
        else:
            return (0, 0), status

    async def getCurrentValue(self, queryId: str) -> ReadRespType:

        result, status = await self.read("getCurrentValue", _queryId=queryId)

        return result, status

    async def getReportingLock(self) -> ReadRespType:

        result, status = await self.read("getReportingLock")

        return result, status

    async def getReporterByTimestamp(self, queryId: str, timestamp: int) -> ReadRespType:

        result, status = await self.read("getReporterByTimestamp", _queryId=queryId, _timestamp=timestamp)

        return result, status

    async def getReporterLastTimestamp(self, reporter: Optional[str] = None) -> ReadRespType:

        result, status = await self.read("getReporterLastTimestamp", _reporter=reporter)

        return result, status

    async def getReportsSubmittedByAddress(self, reporter: Optional[str] = None) -> ReadRespType:

        result, status = await self.read("getReportsSubmittedByAddress", _reporter=reporter)

        return result, status

    async def getReportTimestampByIndex(self, queryId: str, index: int) -> ReadRespType:

        result, status = await self.read("getReportTimestampByIndex", _queryId=queryId, _index=index)

        return result, status

    async def getTimeBasedReward(self) -> Tuple[float, ResponseStatus]:
        result, status = await self.read("getTimeBasedReward")
        if status.ok:
            trb_reward = float(result) / 1.0e18
        else:
            trb_reward = float(0)
        return trb_reward, status

    async def getTimestampCountById(self, queryId: str) -> ReadRespType:

        result, status = await self.read("getTimestampCountById", _queryId=queryId)

        return result, status

    async def getTimeOfLastNewValue(self) -> ReadRespType:

        result, status = await self.read("getTimeOfLastNewValue")

        if status.ok:
            t = TimeStamp(result)
        else:
            t = None

        return t, status

    async def getTimestampIndexByTimestamp(self, queryId: str, timestamp: int) -> ReadRespType:

        result, status = await self.read("getTimestampIndexByTimestamp", _queryId=queryId, _timestamp=timestamp)

        return result, status

    async def getTipsById(self, queryId: str) -> ReadRespType:

        tips, status = await self.read("getTipsById", _queryId=queryId)

        return float(tips) / 1.0e18, status

    async def getTipsByUser(self, user: Optional[str] = None) -> ReadRespType:

        result, status = await self.read("getTipsByUser", _user=user)

        return result, status

    async def getValueByTimestamp(self, queryId: str, timestamp: int) -> ReadRespType:

        result, status = await self.read("getValueByTimestamp", _queryId=queryId, _timestamp=timestamp)

        return result, status

    async def verify(self) -> ReadRespType:

        result, status = await self.read("verify")

        return result, status



""" Test application utilities

"""
import logging
import threading
import time

from telliot_core.apps.app import ThreadedApplication

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


def test_threaded_application():
    """Test application startup and shutdown"""

    app = ThreadedApplication()
    assert isinstance(app._shutdown, threading.Event)

    app = ThreadedApplication()
    app.startup()
    time.sleep(2)
    app.shutdown()



"""telliot_core.apps.config module"""
import json
import logging
from datetime import datetime
from pathlib import Path
from typing import Literal
from typing import Optional
from typing import Type
from typing import Union

import yaml
from clamfig import deserialize
from clamfig import Serializable
from clamfig import serialize
from yaml import Dumper as Dumper
from yaml import Loader as Loader

from telliot_core.utils.home import telliot_homedir

logger = logging.getLogger(__name__)

config_formats = Literal["yaml", "json"]


class ConfigOptions(Serializable):
    """An object used to manage configuration options

    Each attribute represents a configuration option.
    Subclasses should add configuration attributes as required.
    """

    pass


class ConfigFile:
    """Configuration File

    Provides the ability to load and store a `ConfigOptions` object from a
    human readable file in JSON or YAML formats.
    """

    @property
    def config_file(self) -> Path:
        return self.config_dir / f"{self.name}.{self.config_format}"

    def __init__(
        self,
        name: str,
        config_type: Type[ConfigOptions],
        config_dir: Optional[Union[str, Path]] = None,
        config_format: config_formats = "yaml",
    ) -> None:
        """Construct a new ConfigFile object

        Args:
            name:
                Config filename

            config_dir:
                Config file folder  If None, home directory is automatically computed.

            config_format:
                Format of config file ('yaml' or 'json')

        """

        #: Configuration Name
        self.name = name

        #: Config Type
        self.config_type = config_type

        #: Configuration Folder
        self.config_dir = telliot_homedir(config_dir)

        #: File Format
        self.config_format = config_format

        # Try to load configuration from file
        if self.config_file.exists():

            # Try to load file
            _ = self.get_config()

        else:
            # Create a default configuration and save it
            options = self.config_type()
            self.save_config(options)

    def get_config(self) -> ConfigOptions:
        """Load Configuration from a .yaml file"""

        if self.config_format == "yaml":

            with open(self.config_file, "r") as f:
                try:
                    state = yaml.load(f, Loader=Loader)
                    config = deserialize(state)
                except Exception:
                    raise Exception(f"Error reading config file {self.config_file}")

        elif self.config_format == "json":

            with open(self.config_file, "r") as f:
                try:
                    state = json.load(f)
                    config = deserialize(state)
                except Exception:
                    raise Exception(f"Error reading config file {self.config_file}")

        else:
            raise AttributeError(f"Invalid config file type: {self.config_format}")

        return config  # type: ignore

    def save_config(self, config: ConfigOptions) -> None:

        """Save configuration to file"""

        try:

            # Make sure folder exists
            self.config_file.parent.mkdir(parents=True, exist_ok=True)

            # Back up existing file
            if self.config_file.exists():
                dt_str = datetime.now().strftime("%Y%M%d-%H%M%S")
                path = self.config_file
                backup_file = path.with_name(path.stem + "_" + dt_str + path.suffix + ".bak")
                self.config_file.replace(backup_file)

            if self.config_format == "yaml":

                with open(self.config_file, "w") as f:

                    state = serialize(config)
                    yaml.dump(state, f, Dumper=Dumper, sort_keys=False)

            elif self.config_format == "json":

                with open(self.config_file, "w") as f:
                    state = serialize(config)
                    jstr = json.dumps(state, indent=2)
                    f.write(jstr)

            logger.info("Saved config '{}' to {}".format(self.name, self.config_file))

        except FileNotFoundError as e:
            logger.error("Error saving {} to {}".format(self.__class__.__name__, self.config_file))
            raise e



from telliot_core.asset_registry import asset_registry
from telliot_core.model.tokens import BlockChainAsset


def test_asset_registry():
    btc = asset_registry.get("btc")
    assert isinstance(btc, BlockChainAsset)



""" Telliot registry module"""
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple

from clamfig import Serializable

ModelStateType = List[Tuple[str, Dict[str, Any]]]


class RegisteredModel(Serializable):
    pass



from collections import namedtuple
from dataclasses import dataclass
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from telliot_core.model.asset import Asset
from telliot_core.model.base import Base


@dataclass
class BlockChainAsset(Asset):
    symbol: str

    def __post_init__(self) -> None:
        self.symbol = self.symbol.lower()


@dataclass
class ERC20Token(BlockChainAsset):
    """Representation of an ERC20 token"""

    #: Chain ID
    chain_id: int

    #: Contract address on chain
    address: str

    #: Conversion factor in currency calculations
    decimals: int

    #: Logo URI
    logo_uri: str = ""


TokenListVersion = namedtuple("TokenListVersion", ["major", "minor", "patch"], defaults=[0, 0, 0])


@dataclass
class ERC20TokenList(Base):
    """ERC-20 Token List"""

    #: Token list name
    name: str

    #: Token list
    tokens: List[ERC20Token]

    version: Optional[TokenListVersion]

    @classmethod
    def from_uniswap(cls, jsn: Dict[str, Any]) -> "ERC20TokenList":
        """Create a token list from uniswap format

        See `https://tokenlists.org/` for more information
        """

        name = jsn["name"]
        version = TokenListVersion(
            major=jsn["version"]["major"],
            minor=jsn["version"]["minor"],
            patch=jsn["version"]["patch"],
        )

        jtokens = jsn.get("tokens", [])
        if not jtokens:
            raise Exception("Token List is empty")

        tokens = []
        for token in jtokens:
            tokens.append(
                ERC20Token(
                    id=token.get("symbol", None),
                    name=token.get("name", None),
                    symbol=token.get("symbol", None),
                    chain_id=token.get("chainId", None),
                    address=token.get("address", None),
                    decimals=token.get("decimals", None),
                    logo_uri=token.get("logoURI", None),
                )
            )

        return cls(name=name, tokens=tokens, version=version)



"""
Test covering gas price fetching utils
"""
import pytest
from requests.exceptions import SSLError

from telliot_core.gas.legacy_gas import fetch_gas_price


@pytest.mark.asyncio
async def test_fetch_legacy_gas_price(monkeypatch, caplog):
    def raise_ssl_error(*args, **kwargs):
        raise SSLError

    def raise_exception():
        raise Exception

    monkeypatch.setattr("requests.get", lambda x: raise_ssl_error())
    gp = await fetch_gas_price()

    assert gp is None
    assert "SSLError -- Unable to fetch gas price" in caplog.text

    monkeypatch.setattr("requests.get", lambda x: raise_exception())
    gp = await fetch_gas_price()

    assert gp is None
    assert "Error fetching gas price:" in caplog.text



"""
Utils for connecting to an EVM contract
"""
import asyncio
import logging
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

from chained_accounts import ChainedAccount
from eth_typing.evm import ChecksumAddress
from web3 import Web3
from web3.datastructures import AttributeDict

from telliot_core.model.endpoints import RPCEndpoint
from telliot_core.utils.key_helpers import lazy_key_getter
from telliot_core.utils.response import error_status
from telliot_core.utils.response import ResponseStatus

logger = logging.getLogger(__name__)


class Contract:
    """Convenience wrapper for connecting to an Ethereum contract"""

    def __init__(
        self,
        address: Union[str, ChecksumAddress],
        abi: Union[List[Dict[str, Any]], str],
        node: RPCEndpoint,
        account: Optional[ChainedAccount] = None,
    ):

        self.address = Web3.toChecksumAddress(address)
        self.abi = abi
        self.node = node
        self.contract = None
        self.account = account
        self._private_key: Optional[bytes] = None

    def connect(self) -> ResponseStatus:
        """Connect to EVM contract through an RPC Endpoint"""

        if not self.node.web3:
            msg = "node is not instantiated"
            return ResponseStatus(ok=False, error=msg)

        self.node.connect()
        self.contract = self.node.web3.eth.contract(address=self.address, abi=self.abi)
        return ResponseStatus(ok=True)

    async def read(self, func_name: str, *args: Any, **kwargs: Any) -> Tuple[Any, ResponseStatus]:
        """
        Reads data from contract
        inputs:
        func_name (str): name of contract function to call

        returns:
        ResponseStatus: standard response for contract data
        """

        if self.contract:
            try:
                contract_function = self.contract.get_function_by_name(func_name)
                output = contract_function(*args, **kwargs).call()
                return output, ResponseStatus(ok=True)
            except ValueError as e:
                msg = f"function '{func_name}' not found in contract abi"
                return None, ResponseStatus(ok=False, e=e, error=msg)
            except asyncio.exceptions.TimeoutError as e:
                msg = "timeout reading from contract"
                return None, ResponseStatus(ok=False, e=e, error=msg)
            except Exception as e:
                msg = "error reading from contract"
                return None, ResponseStatus(ok=False, e=e, error=msg)
        else:
            msg = "no instance of contract"
            return None, ResponseStatus(ok=False, error=msg)

    @property
    def private_key(self) -> bytes:

        if not self._private_key:
            self._private_key = lazy_key_getter(self.account)

        return self._private_key

    async def write(
        self,
        func_name: str,
        gas_limit: int,
        legacy_gas_price: Optional[int] = None,
        max_priority_fee_per_gas: Optional[int] = None,
        max_fee_per_gas: Optional[int] = None,
        acc_nonce: Optional[int] = None,
        **kwargs: Any,
    ) -> Tuple[Optional[AttributeDict[Any, Any]], ResponseStatus]:
        """For submitting any contract transaction once without retries

        gas price measured in gwei

        """

        # Validate inputs
        if (legacy_gas_price is not None) and ((max_fee_per_gas is not None) or (max_priority_fee_per_gas is not None)):
            raise ValueError(
                """invalid combination of legacy gas arguments
                 and EIP-1559 gas arguments"""
            )

        if (legacy_gas_price is None) and (max_fee_per_gas is None) and (max_priority_fee_per_gas is None):
            raise ValueError("no gas strategy selected!")

        status = ResponseStatus()

        if not acc_nonce:
            acc = self.node.web3.eth.account.from_key(self.private_key)
            acc_nonce = self.node.web3.eth.get_transaction_count(acc.address)

        if not self.contract:
            msg = f"Contract.write({func_name}) error: Unable to connect to contract"
            return None, error_status(msg, log=logger.error)

        if not self.node:
            msg = f"Contract.write({func_name}) error: No node instance"
            return None, error_status(msg, log=logger.error)

        if self.private_key:
            acc = self.node.web3.eth.account.from_key(self.private_key)
        else:
            msg = f"Contract.write({func_name}) error: Private key missing"
            return None, error_status(msg, log=logger.error)

        try:
            # build transaction
            contract_function = self.contract.get_function_by_name(func_name)
            transaction = contract_function(**kwargs)

            # start tx dict with static elements
            tx_dict = {
                "from": acc.address,
                "nonce": acc_nonce,
                "gas": gas_limit,
                "chainId": self.node.chain_id,
            }

            # use legacy gas strategy if only legacy gas price is provided
            if legacy_gas_price is not None:

                if (max_fee_per_gas is not None) or (max_priority_fee_per_gas is not None):
                    raise ValueError(
                        """"cannot use both legacy gas arguments
                        and type 2 transaction (EIP1559) args in one transaction"""
                    )

                tx_dict["gasPrice"] = self.node.web3.toWei(legacy_gas_price, "gwei")

            # use EIP-1559 gas strategy if maxFeePerGas
            # and/or MaxPriorityFeePerGas are provided
            else:
                if legacy_gas_price is not None:
                    raise ValueError(
                        """"cannot use both legacy gas arguments
                         and type 2 transaction (EIP1559) args in one transaction"""
                    )

                if max_fee_per_gas is not None:
                    tx_dict["maxFeePerGas"] = self.node.web3.toWei(max_fee_per_gas, "gwei")

                    if max_priority_fee_per_gas is not None:
                        tx_dict["maxPriorityFeePerGas"] = self.node.web3.toWei(max_priority_fee_per_gas, "gwei")

                    # else if (if legacy price and max fee are not provided)
                    # use max priority fee
                    elif max_priority_fee_per_gas is not None:
                        tx_dict["maxPriorityFeePerGas"] = self.node.web3.toWei(max_priority_fee_per_gas, "gwei")

                    # raise ValueError if no gas arguments are provided
                    else:
                        raise ValueError(
                            """no gas strategy selected!
                            must provide either legacy
                            or EIP-1559 gas arguments"""
                        )
            # pass in tx dict to build the transaction
            built_tx = transaction.buildTransaction(tx_dict)
            # submit transaction
            tx_signed = acc.sign_transaction(built_tx)

        except Exception as e:
            note = "Failed to build transaction"
            return None, error_status(note, log=logger.error, e=e)

        try:
            logger.debug(f"Sending transaction: {func_name}")
            tx_hash = self.node.web3.eth.send_raw_transaction(tx_signed.rawTransaction)

        except Exception as e:
            note = "Send transaction failed"
            return None, error_status(note, log=logger.error, e=e)

        try:
            # Confirm transaction
            tx_receipt = self.node.web3.eth.wait_for_transaction_receipt(tx_hash, timeout=360)

            tx_url = f"{self.node.explorer}/tx/{tx_hash.hex()}"

            if tx_receipt["status"] == 1:
                logger.info(f"{func_name} transaction succeeded. ({tx_url})")
                return tx_receipt, status

            elif tx_receipt["status"] == 0:
                msg = f"{func_name} transaction reverted. ({tx_url})"
                return tx_receipt, error_status(msg, log=logger.error)

            return tx_receipt, status

        except Exception as e:
            note = "Failed to confirm transaction"
            return None, error_status(note, log=logger.error, e=e)



"""
Test covering Pytelliot EVM contract connection utils.
"""
import pytest
import web3

from telliot_core.apps.core import TelliotCore


@pytest.mark.asyncio
async def test_connect_to_tellor(rinkeby_test_cfg):
    """Contract object should access Tellor functions"""
    async with TelliotCore(config=rinkeby_test_cfg) as core:
        tellorx = core.get_tellorx_contracts()
        assert len(tellorx.master.contract.all_functions()) > 0
        assert isinstance(
            tellorx.master.contract.all_functions()[0],
            web3.contract.ContractFunction,
        )


@pytest.mark.asyncio
async def test_mixed_gas_inputs(rinkeby_test_cfg):
    """Contract.write() should refuse a combination of
    legacy gas args and EIP-1559 gas args"""

    with pytest.raises(ValueError):

        async with TelliotCore(config=rinkeby_test_cfg) as core:
            tellorx = core.get_tellorx_contracts()

            tx_receipt, status = await tellorx.oracle.write(
                func_name="transfer",
                _to="0xF90cd1D6C1da49CE2cF5C39f82999D7145aa66aD",
                _amount=1,
                extra_gas_price=0,
                gas_limit=350000,
                legacy_gas_price=1,
                max_fee_per_gas=2,
            )



from telliot_core.model.chain import Chain
from telliot_core.model.chain import ChainList


def test_chain():
    """Test ChainList"""

    cl = ChainList()
    ch = cl.get_chain("ETH", "mainnet")
    assert isinstance(ch, Chain)
    assert ch.chain == "ETH"
    assert ch.chain_id == 1
    assert ch.network == "mainnet"


def test_matic():
    cl = ChainList()
    ch = cl.get_chain("MATIC", "mainnet")
    assert isinstance(ch, Chain)
    assert ch.chain_id == 137

    ch = cl.get_chain("MATIC", "testnet")
    assert ch.chain_id == 80001



from datetime import datetime
from typing import Tuple

import click

from telliot_core.cli.utils import async_run
from telliot_core.cli.utils import cli_core
from telliot_core.utils.timestamp import TimeStamp


@click.group()
def read() -> None:
    """Read on-chain TellorX contracts."""
    pass


# ------------------------------------------------------------------
# Oracle Contract
# ------------------------------------------------------------------
@read.group()
def oracle() -> None:
    """Read from the TellorX oracle contract."""
    pass


@oracle.command()
@click.pass_context
@async_run
async def gettimebasedreward(ctx: click.Context) -> None:
    async with cli_core(ctx) as core:

        tellorx = core.get_tellorx_contracts()
        result, status = await tellorx.oracle.getTimeBasedReward()

        if not status.ok:
            print(status)
        else:
            print(f"{result} TRB")


@oracle.command()
@click.option("--address", type=str, help="Reporter address (starting with 0x).")
@click.pass_context
@async_run
async def getreporterlasttimestamp(ctx: click.Context, address: str) -> None:
    async with cli_core(ctx) as core:

        tellorx = core.get_tellorx_contracts()
        ts, status = await tellorx.oracle.getReporterLastTimestamp(address)

        if not status.ok:
            print(status)
        else:
            print(f"{ts} ({datetime.fromtimestamp(ts)})")


# ------------------------------------------------------------------
# Master Contract
# ------------------------------------------------------------------
@read.group()
def master() -> None:
    """Read from the TellorX master contract."""
    pass


async def get_staker_info(ctx: click.Context, address: str) -> Tuple[str, TimeStamp]:
    """Get staker information."""
    async with cli_core(ctx) as core:
        if not address:
            address = core.get_account().address

        tellorx = core.get_tellorx_contracts()
        (staker_status, date_staked), status = await tellorx.master.getStakerInfo(address=address)
        return staker_status, date_staked


@master.command()
@click.argument("address", required=False)
@click.pass_context
@async_run
async def getstakerinfo(ctx: click.Context, address: str) -> None:
    """Get staker information."""
    (staker_status, date_staked) = await get_staker_info(ctx, address)

    print(f"Status: {staker_status}")
    if staker_status != "NotStaked":
        print(f"Staked on {date_staked} ({date_staked.age} ago)")


@master.command()
@click.argument("dispute_id", type=int, required=True)
@click.pass_context
@async_run
async def disputesbyid(ctx: click.Context, dispute_id: int) -> None:
    """Get disputes by ID."""

    async with cli_core(ctx) as core:
        tellorx = core.get_tellorx_contracts()
        result, read_response = await tellorx.master.disputesById(dispute_id)

    if not read_response.ok:
        click.echo(read_response)
    else:
        click.echo(result)



from dataclasses import dataclass
from dataclasses import field
from typing import List
from typing import Optional

from telliot_core.apps.config import ConfigOptions
from telliot_core.model.base import Base


@dataclass
class EVMCurrency(Base):
    name: str
    symbol: str
    decimals: int


@dataclass
class Chain(Base):
    name: str
    chain: str
    network: str
    chain_id: int
    currency: EVMCurrency


default_chain_list = [
    Chain(
        chain_id=1,
        name="Ethereum Mainnet",
        chain="ETH",
        network="mainnet",
        currency=EVMCurrency(name="Ether", symbol="ETH", decimals=18),
    ),
    Chain(
        chain_id=3,
        name="Ethereum Testnet Ropsten",
        chain="ETH",
        network="ropsten",
        currency=EVMCurrency(name="Ropsten Ether", symbol="ROP", decimals=18),
    ),
    Chain(
        chain_id=4,
        name="Ethereum Testnet Rinkeby",
        chain="ETH",
        network="rinkeby",
        currency=EVMCurrency(name="Rinkeby Ether", symbol="RIN", decimals=18),
    ),
    Chain(
        chain_id=137,
        name="Matic(Polygon) Mainnet",
        chain="Matic(Polygon)",
        network="mainnet",
        currency=EVMCurrency(name="Matic", symbol="MATIC", decimals=18),
    ),
    Chain(
        chain_id=80001,
        name="Matic(Polygon) Testnet Mumbai",
        chain="Matic(Polygon)",
        network="testnet",
        currency=EVMCurrency(name="Matic", symbol="tMATIC", decimals=18),
    ),
    Chain(
        chain_id=69,
        name="Optimism Kovan",
        chain="Optimism",
        network="testnet (public)",
        currency=EVMCurrency(name="Ether", symbol="ETH", decimals=18),
    ),
    Chain(
        chain_id=1666600000,
        name="Harmony Mainnet",
        chain="Harmony",
        network="mainnet",
        currency=EVMCurrency(name="Harmony ONE", symbol="ONE", decimals=18),
    ),
    Chain(
        chain_id=1666700000,
        name="Harmony Testnet",
        chain="Harmony",
        network="testnet",
        currency=EVMCurrency(name="Harmony ONE", symbol="ONE", decimals=18),
    ),
    Chain(
        chain_id=421611,
        name="Arbitrum Rinkeby",
        chain="Arbitrum",
        network="testnet",
        currency=EVMCurrency(name="Ether", symbol="ETH", decimals=18),
    ),
    Chain(
        chain_id=941,
        name="Pulsechain Testnet",
        chain="Pulechain",
        network="testnet",
        currency=EVMCurrency(name="Pulsechain tPLS", symbol="tPLS", decimals=18),
    ),
    Chain(
        chain_id=42161,
        name="Arbitrum One",
        chain="Arbitrum",
        network="mainnet",
        currency=EVMCurrency(name="Arbitrum Ether", symbol="ETH", decimals=18),
    ),
]


@dataclass
class ChainList(ConfigOptions):
    chains: List[Chain] = field(default_factory=lambda: default_chain_list)

    def get_chain(self, chain: str = "ETH", network: str = "rinkeby") -> Optional[Chain]:
        """Get chain"""

        for ch in self.chains:
            if chain.lower() in ch.chain.lower():
                if network.lower() in ch.network.lower():
                    return ch

        return None



import asyncio

import click

from telliot_core.cli.utils import async_run
from telliot_core.cli.utils import cli_core
from telliot_core.contract.listener import event_logger
from telliot_core.directory import contract_directory


@click.command()
@click.pass_context
@async_run
async def listen(ctx: click.Context) -> None:
    """Listen for Tellor network events."""

    async with cli_core(ctx) as core:

        chain_id = core.config.main.chain_id

        if chain_id in [1, 4]:

            master = contract_directory.find(name="tellorx-master", chain_id=chain_id)[0]
            oracle = contract_directory.find(name="tellorx-oracle", chain_id=chain_id)[0]

        # elif chain_id in [137, 80001]:
        #     oracle = contract_directory.find(name='tellorflex-oracle', chain_id=chain_id)[0]

        else:
            click.echo(f"Listening not supported on network: {chain_id}")
            return

        # await core.listener.subscribe_new_blocks(handler=block_logger)

        # Subscribe to contract events
        await core.listener.subscribe_contract_events(
            handler=event_logger,
            address=master.address[core.config.main.chain_id],
        )
        await core.listener.subscribe_contract_events(
            handler=event_logger,
            address=oracle.address[core.config.main.chain_id],
        )

        try:
            while True:
                await asyncio.sleep(1)

        except asyncio.exceptions.CancelledError:
            core.log.debug("CLI Listener cancelled.")



from datetime import datetime

from telliot_core.utils.timestamp import TimeStamp


def test_main():
    """Test the TimeStamp class"""
    dt = datetime(1999, 12, 31, 12, 59, 59)
    dt_ts = TimeStamp(dt.timestamp())

    print(dt_ts.dt)

    d = datetime(1999, 12, 31)
    d_ts = TimeStamp(d.timestamp())
    print(d_ts.dt)

    print(d_ts.age)


def test_repr():
    """Test the Timestamp class __repr__ method"""
    ts = 1639756400
    ts = TimeStamp(ts)
    assert repr(ts) == "TimeStamp(1639756400)"



"""
Tests covering Pytelliot rpc connection  utils.
"""
import pytest
import requests
from brownie import chain

from telliot_core.model.endpoints import EndpointList
from telliot_core.model.endpoints import RPCEndpoint

network = "mainnet"
provider = "pokt"


def test_rpc_endpoint():
    """RPCEndpoint connects to the blockchain"""
    url = "http://127.0.0.1:8545"  # local Ganache node
    endpt = RPCEndpoint(network=network, provider=provider, url=url)
    endpt.connect()
    chain.mine(10)
    assert endpt.web3.eth.block_number > 1

    print(endpt.get_state())


def test_very_bad_rpc_url():
    """an invalid url will raise an exception in RPCEndpoint"""
    url = "this is not a valid rpc url"
    endpt = RPCEndpoint(network=network, provider=provider, url=url)
    with pytest.raises(ValueError):
        _ = endpt.connect()


def test_incomplete_rpc_url():
    """an incomplete url will raise an exception in RPCEndpoint"""
    url = "https://eth-rinkeby.gateway.pokt.network/v1/lb/"
    endpt = RPCEndpoint(network=network, provider=provider, url=url)
    # expect bad url error from requests library
    with pytest.raises(requests.exceptions.HTTPError):
        _ = endpt.connect()


def test_endpoint_list():
    sl = EndpointList()
    # print(json.dumps(sl.get_state(), indent=2))
    ep4 = sl.find(chain_id=4)[0]
    assert ep4.network == "rinkeby"

    ep3 = sl.find(chain_id=3)[0]
    assert ep3.network == "ropsten"
